(* 
  Coq code for the paper:
  Interfacing Hoare Logic and Type Systems for Foundational Proof-Carrying Code
  by Nadeem Abdul Hamid and Zhong Shao

  Coq version 8.0

  This code is a prototype version only -- proofs are mostly complete.
*)

(*
 *
 Array-Extensible-TAL-with-polymorphism

 A version of TAL supporting primitive array types
 but no operations on them!

 A revised version of (atal.v) separating code and data heaps

 For Coq Version 8.0
*)


(* Use this to fudge through some proofs :) *)
Axiom Super : forall P:Prop, P.


Require Import Eqdep.
Require Import Arith.
Require Import ArithRing.
Require Import EqNat.
Require Import Omega.
Require Import mystuff.
Require Import tis.
Require Import captis.
Require Import fmap.

Require Import axtalp.

(**************************************************************************)
(**************************************************************************)
(**************************************************************************)
(*************************************************************************)
(*************************************************************************)
(*************************************************************************)
(* TRANSLATION TO CAP *)

Require Import talreg.

Definition reg2regt (r : Reg) : Regt := nat2regt (reg2nat r).
Coercion reg2regt : Reg >-> Regt.

Section iXTALtoTIS.


(* library code area memory *)
Definition ExtCodeTy := Word -> optT (list Cmd ** Pred).
Variable ExtCode : ExtCodeTy.

(* code layout function, mapping from iXTAL code labels to
   TIS machine address words *)
Definition LytType := fmap int Word.
Definition Lytlook := fmaplook (A:=int) (B:=Word).

Variable CLyt : LytType.  (* for CHeap *)
Variable DLyt : LytType.  (* for Heap *)

(* utilities *)
Definition MemPred := Mem -> Prop.

(*************************************************************************)
Fixpoint InstrSeq_size (Is : InstrSeq) : int :=
  match Is with
  | iseq _ Is' => S (InstrSeq_size Is')
  | _ => 1
  end.

(* [w1,...,wn] => M@w *)
Fixpoint WordsInMem (ws : list Word) (w : Word) {struct ws} : MemPred :=
  match ws with
  | nil => fun _ => True
  | w' :: ws' => fun M => M w = w' /\ WordsInMem ws' (S w) M
  end.

Definition CmdsInMem (Cs : list Cmd) w Mp : Prop :=
  exists Ws, map Dc Ws = Cs /\ WordsInMem Ws w Mp.

Fixpoint tailnth (n : int) A (Ls : list A) {struct n} : 
 option (list A) :=
  match n with
  | O => Some Ls
  | S m => match Ls with
           | nil => None
           | _ => tailnth m _ (tail Ls)
           end
  end.

(*************************************************************************)
(* Translation of programs *)
Definition TrWordVal (wv : WordVal) (w : Word) : Prop :=
  match wv with
  | wl l => Lytlook CLyt l w \/ Lytlook DLyt l w
  | wi i => w = i
  end.

Inductive TrInstr : list Cmd -> Instr -> list Cmd -> Prop :=
  | add_tri :
      forall Cs rd rs rt, TrInstr Cs (add rd rs rt) (_add rd rs rt :: Cs)
  | movi_tri : forall Cs rd i, TrInstr Cs (movi rd i) (_movi rd i :: Cs)
  | movl_tri :
      forall Cs rd l w,
      Lytlook CLyt l w -> TrInstr Cs (movl rd l) (_movi rd w :: Cs).

Inductive TrInstrSeq : InstrSeq -> list Cmd -> Prop :=
  | iseq_tris :
      forall i Is cs Cs,
      TrInstrSeq Is Cs -> TrInstr Cs i cs -> TrInstrSeq (iseq i Is) cs
  | jd_tris :
      forall l Ts w, Lytlook CLyt l w -> TrInstrSeq (jd l Ts) (_jd w :: nil)
  | jmp_tris : forall r Ts, TrInstrSeq (jmp r Ts) (_jmp r :: nil).

Inductive TrCHeapVal : CHeapVal -> WordList -> Prop :=
  | tr_codehval :
      forall Is T Ws, TrInstrSeq Is (map Dc Ws) -> TrCHeapVal (code T Is) Ws
  | tr_stubhval :
      forall T, (* stubs handled at top level *)
      TrCHeapVal (stub T) nil.

Inductive TrHeapVal : HeapVal -> WordList -> Prop :=
    tr_arrayhval :
      forall n Ns, n = length Ns -> TrHeapVal (array Ns) (n :: Ns).

Definition TrRegFile (R : RegFile) (Rt : RFile) : Prop :=
  forall r v w, R r = v -> Rt r = w -> TrWordVal v w.

Definition TrPC (CH : CHeap) (Is : InstrSeq) (pc : Word) : Prop :=
  exists l,
    (exists T,
       (exists Is',
          (exists w,
             (exists n,
                fmaplook CH l (code T Is') /\
                SubseqIS Is Is' n /\ Lytlook CLyt l w /\ pc = w + n)))).

Definition TrCHeap (CH : CHeap) (M : Mem) : Prop :=
  forall l chv,
  fmaplook CH l chv ->
  exists Ws,
    (exists a, Lytlook CLyt l a /\ TrCHeapVal chv Ws /\ WordsInMem Ws a M).

Definition rAlloc := rP.

(* DEFINITION OF DATA LYT FUNCTION CONSISTENCY (non-overlapping/within 
     allocation region) IS EMBEDDED HERE: *)
Definition TrHeap (H : Heap) (M : Mem) (R : RFile) : Prop :=
  forall l hv,
  fmaplook H l hv ->
  exists Ws,
    (exists a,
       Lytlook DLyt l a /\
       TrHeapVal hv Ws /\
       WordsInMem Ws a M /\
       a + length Ws < R rAlloc /\
       (forall l' hv' a' Ws',
        fmaplook H l' hv' ->
        l <> l' ->
        Lytlook DLyt l' a' ->
        TrHeapVal hv' Ws' -> a + length Ws < a' \/ a' + length Ws' < a)).

Definition ExtCLytConsistent (ExtCode : ExtCodeTy) 
  (CH : CHeap) :=
  forall a w G Is,
  Lytlook CLyt a w -> fmaplook CH a (code G Is) -> ExtCode w = noneT _.

Definition ExtLytPred (CH : CHeap) : Prop :=
  onetoonef CLyt /\ eqdomf _ _ _ CLyt CH /\ disjf CLyt DLyt.

Inductive TrProgram : Program -> State -> Prop :=
    trprogram_c :
      forall CH H R Is MM RR pc,
      TrCHeap CH MM ->
      TrHeap H MM RR ->
      TrRegFile R RR ->
      TrPC CH Is pc ->
      ExtLytPred CH ->
      ExtCLytConsistent ExtCode CH ->
      (forall w Cs Pcs, ExtCode w = someT (Cs,* Pcs) -> CmdsInMem Cs w MM) ->
      TrProgram (CH, (H, (R, Is))) (MM, (RR, pc)).

End iXTALtoTIS.

(*************************************************************************)
(*************************************************************************)
(*************************************************************************)
(* NOW THE PROOFS OF iXTAL-CAP TRANSLATION WELL-FORMEDNESS *)

Section Proofs.

Variable ExtCode : ExtCodeTy.
Variable CLyt : LytType.

Definition TrivSP (St : State) := True.

Definition CpInv (CH : CHeap) (T : CodeTy) : Pred :=
  fun St : State =>
  exists DLyt,
    (exists H,
       (exists R,
          (exists HT,
             (exists Ts,
                (exists G,
                   match St with
                   | (MM, (RR, _)) =>
                       InstCodeTy T Ts G /\
                       WFCHeap CH /\
                       WFHeap H HT /\
                       WFRegFile CH HT R G /\
                       TrCHeap CLyt CH MM /\
                       TrHeap DLyt H MM RR /\
                       TrRegFile CLyt DLyt R RR /\ ExtLytPred CLyt DLyt CH
                   end))))).

Definition CpInterf := CpInv.

(* what it means for a CAP code specification to have been properly
   generated from an iXTAL code heap type and external code instantiation *)
Definition IsCpGen (CT : CdSpec) (CH : CHeap) : Prop :=
  (* extcode preconditions are directly used for CT *)
  (forall w Cs P, 
      ExtCode w = someT (Cs,* P) -> CT w = someT (length Cs,* P)) 
  /\
  (* iXTAL code types are used thru CpInv for CT *)
  (forall l w T Is,
   fmaplook CH l (code T Is) ->
   Lytlook CLyt l w -> CT w = someT (InstrSeq_size Is,* CpInv CH T)) 
  /\
  (* for all iXTAL stubs, there must be the corresponding external code *)
  (forall l w T,
   fmaplook CH l (stub T) ->
   Lytlook CLyt l w ->
   exists Cs, (exists P, ExtCode w = someT (Cs,* P))) 
  (* for all CAP preconditions that are in CT, either it comes
        from the extcode preconditions,
        or from the iXTAL code block type *)
  (* interface correctness is also checked here - i.e. that if
        ExtCode(CLyt(l)) = P and CH(l)=G then
           (forall St, (CpInv .. G St) -> (P St)) *)
  /\
  (forall w n P,
   CT w = someT (n,* P) ->
   (exists Cs,
      ExtCode w = someT (Cs,* P) /\
      n = length Cs /\ 
     ((exists l, Lytlook CLyt l w) -> 
            (forall l G, (Lytlook CLyt l w) -> (* forall G ? or exists G ?? *)
	      (
	          fmaplook CH l (stub G) ->
	          (forall St, CpInterf CH G St -> P St)))))
      \/
   (exists l,
      (exists T,
         (exists Is,
            Lytlook CLyt l w /\
            fmaplook CH l (code T Is) /\
            n = InstrSeq_size Is /\ P = CpInv CH T))))
  (* if l is in the layout, it better have a specification in CT *)
  /\
  (forall l w, Lytlook CLyt l w -> exists PP, CT w = someT PP)
.


(************************************************************************)
(* smaller lemmas *)

Lemma neq_reg_regt :
 forall (r r' : Reg) (ee : r <> r'), reg2regt r <> reg2regt r'.
intros r r'; case r; case r'; intros; try (intro H; discriminate H);
 try Absurdity.
Qed.


Lemma trpc_iseq_S :
 forall CH i Is pc, TrPC CLyt CH (iseq i Is) pc -> TrPC CLyt CH Is (S pc).
unfold TrPC in |- *; intros.
MCElimEx l int; MCElimEx4 T Is' w n CodeTy.
Exists4 l T Is' w; exists (S n); split; auto.
split; auto.
apply SubseqIS_iseq with i; auto.
split; auto.
unfold Word in |- *; unfold Word in w; omega.
Qed.


Lemma cdspec_heapstub_extcode :
 forall CT CH l G w (D : onetoonef CLyt) (D0 : IsCpGen CT CH)
   (D1 : fmaplook CH l (stub G)) (D2 : Lytlook CLyt l w),
 exists Cs,
   (exists P,
      ExtCode w = someT (Cs,* P) /\
      CT w = someT (length Cs,* P) /\ (forall St, CpInterf CH G St -> P St)).
intros.
generalize D0; intros (DA, (DB, (_, (DC, DD)))).
match goal with
| H0:(Lytlook CLyt l w) |- _ =>
    generalize (DD _ _ H0); intros ((n, P), DD1)
end.
elim (DC _ _ _ DD1); intros DD2.

MCElimEx Cs (list Cmd).
match goal with 
| H2:(exists l, Lytlook _ _ _) -> (_) |- _ => 
      generalize (H2 (ex_intro (fun l => fmaplook CLyt l w) l D2) _ G D2 D1);
      IntrosSplitH
end.
Exists2 Cs P; split; auto.

MCElimEx3 l' G' Is' int.
assert (l' = l).
  apply (D l' l w); auto.
MCRewriteIn l' fmaplook.
UnfoldFmaps.
match goal with
| H2:(CH l = Some (code _ _)) |- _ => rewrite D1 in H2; discriminate H2
end.
Qed.


Lemma TrInstrSeq_subseq_tailnth :
 forall Lyt nis Is Is' wpc Ws' MM,
 SubseqIS Is Is' nis ->
 (*    (Lytlook Lyt x wpc) -> *)
 TrInstrSeq Lyt Is' (map Dc Ws') ->
 WordsInMem Ws' wpc MM ->
 exists Ws : WordList,
   tailnth nis Word Ws' = Some Ws /\
   TrInstrSeq Lyt Is (map Dc Ws) /\ WordsInMem Ws (nis + wpc) MM.
induction nis; intros; MCInv SubseqIS.
exists Ws'; split; auto.

MCRewriteIn Is' TrInstrSeq; MCInv TrInstrSeq.
assert (exists _i, _i :: Cs = map Dc Ws').
  MCInv TrInstr; eauto.
MCElimEx _i Cmd.
induction Ws'.
MCInv (_i :: Cs = map Dc nil).
clear IHWs'.

assert (Cs = map Dc Ws');
 [ UnfoldFoldIn map H11; injection H11; auto | auto ].
rewrite H12 in H8.
UnfoldFoldIn WordsInMem H1.
elim (IHnis _ _ _ _ _ H5 H8 H14); IntrosSplitH.
exists x.
split; auto.
split; auto.
replace (S nis + wpc) with (nis + S wpc); auto.
omega.
Qed.




Lemma cmdsin_flatten : forall Cs MM w, CmdsInMem Cs w MM -> Flatten Cs MM w.
  induction Cs; intros.
compute in |- *; auto.
split; auto.
MCElimEx Ws CmdsInMem.
induction  Ws as [| a0 Ws HrecWs].
MCUnfoldIn map eq; MCInv eq.
clear HrecWs.
MCUnfoldIn map eq; MCInv eq.
MCInv WordsInMem.

MCElimEx x (CmdsInMem (a :: Cs) w MM).
induction  x as [| a0 x Hrecx].
MCUnfoldIn map eq; MCInv eq.
clear Hrecx.
MCUnfoldIn map eq; MCInv eq.
MCRewriteRev Cs.
apply IHCs; auto.
exists x; split; auto.
MCUnfoldInSelf WordsInMem; SplitAllHyps.
Qed.

(* *)
Lemma TrInstrSeq_size_eq :
 forall Lyt Is Cs, TrInstrSeq Lyt Is Cs -> length Cs = InstrSeq_size Is.
induction Is; intros; MCInv TrInstrSeq.
MCInv TrInstr; unfold length, InstrSeq_size in |- *;
 fold length InstrSeq_size in |- *; apply eq_S; apply IHIs; 
 auto.
Qed.

(* *)
Lemma tailnth_length_eq :
 forall A n Ls Ls', tailnth n A Ls' = Some Ls -> length Ls' = n + length Ls.
induction n; UnfoldFold tailnth.
intros; MCInv eq.
intros Ls Ls'; case Ls'; intros; try AbsurdEq.
replace (length (a :: l)) with (S (length l));
 [ auto | compute in |- *; auto ].
replace (S n + length Ls) with (S (n + length Ls)); auto.
Qed.

(* *)
Lemma TrInstr_func :
 forall Lyt Cs0 i Cs Cs',
 TrInstr Lyt Cs0 i Cs -> TrInstr Lyt Cs0 i Cs' -> Cs = Cs'.
induction i; intros; inversion H; inversion H0; auto.
repeat MCUnfoldInSelf Lytlook; UnfoldFmaps; MCInjection (Lyt n); MCRewrite w0.
Qed.

(* *)
Lemma TrInstrSeq_func :
 forall Lyt Is Cs Cs',
 TrInstrSeq Lyt Is Cs -> TrInstrSeq Lyt Is Cs' -> Cs = Cs'.
induction Is; intros.
MCInv (TrInstrSeq Lyt (iseq i Is) Cs); MCInv (TrInstrSeq Lyt (iseq i Is) Cs').
generalize (IHIs _ _ H3 H8); intros.
MCRewriteIn Cs1 (TrInstr Lyt Cs1 i Cs').
apply (TrInstr_func _ _ _ _ _ H5 H10).

MCInv (TrInstrSeq Lyt (jd n o) Cs); MCInv (TrInstrSeq Lyt (jd n o) Cs').
repeat MCUnfoldInSelf Lytlook; UnfoldFmaps; MCInjection (Lyt n); MCRewrite w0.

MCInv (TrInstrSeq Lyt (jmp r o) Cs); MCInv (TrInstrSeq Lyt (jmp r o) Cs').
Qed.

(* *)
Lemma SubseqIS_Tr_tailnth :
 forall Lyt n Is Is' Cs Cs',
 TrInstrSeq Lyt Is Cs ->
 TrInstrSeq Lyt Is' Cs' -> SubseqIS Is Is' n -> tailnth n _ Cs' = Some Cs.
induction n.
intros; MCInv SubseqIS.
replace Cs' with Cs; [ compute in |- *; auto | auto ].
MCRewriteIn Is' TrInstrSeq.
apply TrInstrSeq_func with Lyt Is; auto.

intros.
MCInv SubseqIS.
MCRewriteIn Is' (TrInstrSeq Lyt Is' Cs').
MCInv (TrInstrSeq Lyt (iseq i I) Cs').
generalize (IHn _ _ _ _ H H8 H5); intros.
MCInv TrInstr.
Qed.

(* *)
Lemma IsCodeAreaTrPC :
 forall CT CH DLyt MM pc Is Cs Ws (D0 : IsCpGen CT CH)
   (D1 : ExtLytPred CLyt DLyt CH) (D2 : WFCHeap CH) 
   (D4 : TrPC CLyt CH Is pc) (D5 : TrCHeap CLyt CH MM)
   (D6 : TrInstrSeq CLyt Is Cs) (D7 : WordsInMem Ws pc MM)
   (D8 : Cs = map Dc Ws), IsCodeArea CT pc (length Cs).
intros.
MCInv TrPC.
MCElimEx4 T Is' wpc nis CodeTy.
match goal with
| H1:(TrCHeap _ CH _),H2:(fmaplook CH _ _) |- _ =>
    generalize (H1 _ _ H2); intros (Ws', (apc, DD)); SplitAllHyps
end.
generalize D0; intros (_, (DD2, _)).
exists (CpInv CH T).
exists wpc.
exists (InstrSeq_size Is').
exists nis.

split; auto.
apply DD2 with x; auto.

split; auto.
MCRewrite pc.
replace (InstrSeq_size Is') with (length (map Dc Ws'));
 [ auto | MCInv TrCHeapVal; apply TrInstrSeq_size_eq with CLyt; auto ].
MCRewrite Cs.
replace (wpc + nis + length (map Dc Ws)) with
 (wpc + (nis + length (map Dc Ws))); try symmetry  in |- *;
 try apply plus_assoc_reverse; auto.
apply plus_le_compat_l.
replace (nis + length (map Dc Ws)) with (length (map Dc Ws')).
omega.
apply tailnth_length_eq.
apply SubseqIS_Tr_tailnth with CLyt Is Is'; auto.
MCRewrite (map Dc Ws); auto.
MCInv TrCHeapVal.
Qed.

Lemma trinstrseq_subseq :
 forall n Is Is' Cs',
 TrInstrSeq CLyt Is' Cs' ->
 SubseqIS Is Is' n -> exists Cs, TrInstrSeq CLyt Is Cs.
induction n; intros; MCInv SubseqIS.
exists Cs'; auto.

MCRewriteIn Is' TrInstrSeq.
MCInv TrInstrSeq.
apply IHn with I Cs; auto.
Qed.


(*************************************************************************)
(* BIGGER LEMMAS *)
Lemma trpc_eq_dcpc :
 forall CH MM i Is _i pc (D0 : TrCHeap CLyt CH MM)
   (D1 : TrPC CLyt CH (iseq i Is) pc)
   (D3 : forall Cs, TrInstr CLyt Cs i (_i :: Cs)), 
 _i = Dc (MM pc).
intros.
MCElimEx l' TrPC; MCElimEx4 T' Is' w' n' CodeTy.
match goal with
| H:(TrCHeap _ _ _),H1:(fmaplook CH _ _) |- _ =>
    generalize (H _ _ H1); intros (Ws', (a', (DD1, (DD2, DD3))));
     cut (a' = w');
     [ intros DD4; rewrite DD4 in DD1; rewrite DD4 in DD3; MCClear a'
     | auto ]
end.

assert
 (exists Ws'',
    tailnth n' _ Ws' = Some Ws'' /\
    TrInstrSeq CLyt (iseq i Is) (map Dc Ws'') /\ WordsInMem Ws'' (w' + n') MM).
   replace (w' + n') with (n' + w'); try omega.
   apply TrInstrSeq_subseq_tailnth with Is'; auto.
MCInv TrCHeapVal.
MCElimEx Ws'' (list Word).
MCRewriteIn (w' + n') WordsInMem.
assert
 (exists wA,
    (exists WsA,
       (exists CsA, _i :: CsA = map Dc Ws'' /\ Ws'' = wA :: WsA))).
   MCInv TrInstrSeq.
   MCInv TrInstr;
    repeat
     (induction Ws''; try clear IHWs'';
       [ match goal with
         | H22:(_ :: _ = map _ _) |- _ =>
             unfold map in H22; discriminate H22
         end
       | exists a; exists Ws''; exists Cs; split; auto; generalize (D3 Cs);
          MCRewrite i; intros D3'; inversion D3'; auto ]).

replace w0 with w; repeat MCUnfoldInSelf Lytlook; UnfoldFmaps;
 MCInjection (CLyt l).

MCElimEx3 wA WsA CsA Word.
MCRewriteIn Ws'' (_i :: CsA = map Dc Ws'');
 MCUnfoldIn map (_i :: CsA = map Dc (wA :: WsA)).
injection H7; intros.
MCRewriteIn Ws'' WordsInMem; MCUnfoldInSelf WordsInMem; SplitAllHyps.
MCRewrite (MM pc); MCRewrite (Dc wA).

repeat MCUnfoldInSelf Lytlook; UnfoldFmaps; MCInjection (CLyt l').
Qed.



Lemma fpcc_preserv_movi :
 forall DLyt CH H R rd i Is St H' R' Is'
   (D0 : TrProgram ExtCode CLyt DLyt (CH, (H, (R, iseq (movi rd i) Is))) St)
   (D1 : Eval CH (H, (R, iseq (movi rd i) Is)) (H', (R', Is'))),
 TrProgram ExtCode CLyt DLyt (CH, (H', (R', Is'))) (Step St).
intros.
inversion D0.
MCRewriteIn St TrProgram; MCClear2 Is0 St; MCClearAll.
inversion D1.
(*Match Context With [H8:(?)=rd;H9:(?)=i|-?] -> Unfold i0 in H8;
     Unfold i0 in H9.*)
unfold R'0 in |- *; MCRewrite R0.
(*Clear i0;*) MCClear4 Is0 R'0 R0 H0.
MCRewrite rd0; MCClear rd0.
MCRewrite H'; MCRewriteIn H' Eval; MCRewrite Is'; MCRewriteIn Is' Eval;
 MCClear2 Is' CH0; MCRewrite n; MCClear2 H' n.

unfold Step in |- *; replace (Dc (MM pc)) with (_movi rd i); auto.

constructor; auto.

intros l hv DHH.
generalize (H6 _ _ DHH); intros (Ws, (a, DHH')); SplitAllHyps.
Exists2 Ws a; split; auto.
split; auto.
split; auto.
split; auto.
unfold updatereg in |- *.
replace (beq_regt rAlloc rd) with false; auto.
case rd; compute in |- *; auto.

intro r; compare r rd; intro ee.
rewrite ee; unfold rfupd, updatereg in |- *;
 replace (beq_reg rd rd) with true; auto; replace (beq_regt rd rd) with true;
 auto; intros.
MCRewrite v; MCRewrite w; constructor; auto.

unfold rfupd, updatereg in |- *; replace (beq_reg rd r) with false;
 try apply beq_neq_false; auto; replace (beq_regt r rd) with false;
 try apply beq_regt_false_neqR; try eauto;
 try (apply not_eq_sym; apply neq_reg_regt; auto).
 apply neq_reg_regt; auto.

apply trpc_iseq_S with (movi rd i); auto.

apply trpc_eq_dcpc with CH (movi rd i) Is; auto.
intros; constructor; auto.
Qed.


Lemma fpcc_preserv_movl :
 forall DLyt CH H R rd l Is St H' R' Is'
   (D0 : TrProgram ExtCode CLyt DLyt (CH, (H, (R, iseq (movl rd l) Is))) St)
   (D1 : Eval CH (H, (R, iseq (movl rd l) Is)) (H', (R', Is'))),
 TrProgram ExtCode CLyt DLyt (CH, (H', (R', Is'))) (Step St).
intros.
inversion D0.
MCRewriteIn St TrProgram; MCClear2 Is0 St; MCClearAll.
inversion D1.
(*Match Context With [H8:(?)=rd;H9:(?)=l|-?] 
     -> Unfold i in H8; Unfold i in H9.*)
unfold R'0 in |- *; MCRewrite R0.
(*Clear i;*) MCClear4 Is0 R'0 R0 H0.
MCRewrite rd0; MCClear rd0.
MCRewrite H'; MCRewriteIn H' Eval; MCRewrite Is'; MCRewriteIn Is' Eval;
 MCClear2 Is' CH0; MCRewrite l0; MCClear2 H' l0.

assert (exists w, Lytlook CLyt l w).
   MCElimEx l' TrPC; MCElimEx4 T' Is' w' n' CodeTy.
   generalize (H5 (*TrCHeap*)  _ _ H0); intros (Ws', (a', HA')); SplitAllHyps.
   MCInv TrCHeapVal.
   elim (trinstrseq_subseq _ _ _ _ H16 H2); intros Cs; intros HTR.
   inversion HTR.
   MCInv TrInstr; eauto.

MCElimEx w Word; unfold Step in |- *; replace (Dc (MM pc)) with (_movi rd w);
 auto.

constructor; auto.

intros l' hv' DHH.
generalize (H6 _ _ DHH); intros (Ws, (a, DHH')); SplitAllHyps.
Exists2 Ws a; split; auto.
split; auto.
split; auto.
split; auto.
unfold updatereg in |- *.
replace (beq_regt rAlloc rd) with false; auto.
case rd; compute in |- *; auto.

intro r; compare r rd; intro ee.
rewrite ee; unfold rfupd, updatereg in |- *;
 replace (beq_reg rd rd) with true; auto; replace (beq_regt rd rd) with true;
 auto; intros.
MCRewrite v; MCRewrite w0; constructor; auto.

unfold rfupd, updatereg in |- *; replace (beq_reg rd r) with false;
 try apply beq_neq_false; auto; replace (beq_regt r rd) with false;
 try apply beq_regt_false_neqR; try eauto;
 try (apply not_eq_sym; apply neq_reg_regt; auto).
 (apply neq_reg_regt; auto).
apply trpc_iseq_S with (movl rd l); auto.

apply trpc_eq_dcpc with CH (movl rd l) Is; auto.
intros; constructor; auto.
Qed.



Lemma fpcc_preserv_add :
 forall DLyt CH H R rd rs rt Is St H' R' Is'
   (D0 : TrProgram ExtCode CLyt DLyt (CH, (H, (R, iseq (add rd rs rt) Is)))
           St)
   (D1 : Eval CH (H, (R, iseq (add rd rs rt) Is)) (H', (R', Is'))),
 TrProgram ExtCode CLyt DLyt (CH, (H', (R', Is'))) (Step St).
intros.
inversion D0.
MCRewriteIn St TrProgram; MCClear2 Is0 St; MCClearAll.
inversion D1.
(*Match Context With [H8:(?)=rd;H9:(?)=rs;H10:(?)=rt|-?] -> Unfold i in H8;
     Unfold i in H9; Unfold i in H10.*)
unfold R'0 in |- *; MCRewrite R0.
(*Clear i;*) MCClear4 Is0 R'0 R0 H0.
MCRewrite rd0;
  (*MCRewriteIn rs0 '(R rs0)=(wi s); 
      MCRewriteIn rt0 '(R rt0)=(wi t); MCClear3 rd0 rs0 rt0. *)
 MCRewrite H'; MCRewriteIn H' Eval; MCRewrite Is'; 
 MCRewriteIn Is' Eval; MCClear2 Is' CH0.

unfold Step in |- *; replace (Dc (MM pc)) with (_add rd rs rt).

constructor; auto.

intros l' hv' DHH.
generalize (H6 _ _ DHH); intros (Ws, (a, DHH')); SplitAllHyps.
Exists2 Ws a; split; auto.
split; auto.
split; auto.
split; auto.
unfold updatereg in |- *.
replace (beq_regt rAlloc rd) with false; auto.
case rd; compute in |- *; auto.

intro r; compare r rd; intro ee.
rewrite ee; unfold rfupd, updatereg in |- *;
 replace (beq_reg rd rd) with true; auto; replace (beq_regt rd rd) with true;
 auto; intros.
MCRewrite v; MCRewrite w; replace (RR rs) with s;
 [ replace (RR rt) with t | auto ]; auto.
constructor; auto.
match goal with
| H:(TrRegFile _ _ _ _),H5:(R rt = wi t) |- _ =>
    generalize (H rt (wi t) (RR rt) H5 (refl_equal (RR rt))); intros DDH
end.
inversion DDH; auto.
match goal with
| H:(TrRegFile _ _ _ _),H5:(R rs = wi s) |- _ =>
    generalize (H rs (wi s) (RR rs) H5 (refl_equal (RR rs))); intros DDH
end.
inversion DDH; auto.

unfold rfupd, updatereg in |- *; replace (beq_reg rd r) with false;
 try apply beq_neq_false; auto; replace (beq_regt r rd) with false;
 try apply beq_regt_false_neqR; try eauto;
 try (apply neq_reg_regt; auto).

apply trpc_iseq_S with (add rd rs rt); auto.

apply trpc_eq_dcpc with CH (add rd rs rt) Is; auto.
intros; constructor; auto.
Qed.


Lemma fpcc_preserv_jd :
 forall DLyt CH H R l Ts St w T0 Is
   (D0 : TrProgram ExtCode CLyt DLyt (CH, (H, (R, jd l Ts))) St)
   (D1 : fmaplook CH l (code T0 Is)) (D2 : Lytlook CLyt l w)
   (D3 : curcmd St (_jd w)),
 TrProgram ExtCode CLyt DLyt (CH, (H, (R, Is))) (Step St).
intros; inversion D0.
MCRewriteIn St curcmd; MCRewriteIn St TrProgram; MCClear2 Is0 St; MCClearAll.
unfold curcmd, curmem, curpc in D3; unfold Step in |- *; rewrite D3.
constructor; auto.
unfold TrPC in |- *.
Exists4 l T0 Is w; exists 0.
split; auto.
split;
 [ constructor | split; unfold Word in |- *; unfold Word in w; try omega ];
 auto.
Qed.



Lemma cpinv_preserv_add :
 forall CH T rd rs rt Is St (D : CpInv CH (basect T) St)
   (D1 : WFInstrSeq CH T (iseq (add rd rs rt) Is))
   (D2 : curcmd St (_add rd rs rt)),
   CpInv CH (basect (rftupd T rd intty)) (Step St).
intros CH T rd rs rt Is (MM, (RR, pc)); unfold curcmd, curmem, curpc in |- *;
 intros.
ElimEx6 D DLyt H R HT Ts G.
MCInv InstCodeTy.
MCRewriteIn T WFInstrSeq.
match goal with
| H0:(WFCHeap CH),H2:(WFHeap H HT),H1:(WFRegFile CH HT R G) |- _ =>
    generalize (progress_add _ _ _ _ _ _ _ _ _ H0 H2 H1 D1);
     intros ((H'', (R'', Is'')), DDP');
     elim (preserv_add _ _ _ _ _ _ _ _ _ _ _ _ H0 H2 H1 D1 DDP'); 
     intros HT'; IntrosSplitH
end.

unfold Step in |- *; replace (Dc (MM pc)) with (_add rd rs rt); auto.
Exists4 DLyt H'' R'' HT'.
Exists2 (nil (A:=Omega)) (rftupd G rd intty); split; auto.
constructor; auto.
split; auto.
split; auto.
split; auto.
split; auto.
MCInv Eval.
MCRewrite H''; unfold R' in |- *; MCRewrite R0; MCClear2 H'' R'; auto.
split; auto.

intros l hv DHH.
generalize (H5 _ _ DHH); intros (Ws, (a, DHH')); SplitAllHyps.
Exists2 Ws a; split; auto.
split; auto.
split; auto.
split; auto.
unfold updatereg in |- *.
replace (beq_regt rAlloc rd) with false; auto.
case rd; compute in |- *; auto.

split; auto.
(*Match Context With [H15:(?)=rd; H16:(?)=rs; H17:(?)=rt|-?]
 -> (Unfold i in H15; Unfold i in H16; Unfold i in H17).*)
MCRewrite rd0.

intro r; compare r rd; intro ee.
rewrite ee; unfold rfupd, updatereg in |- *;
 replace (beq_reg rd rd) with true; auto; replace (beq_regt rd rd) with true;
 auto; intros.
MCRewrite v; MCRewrite w; replace (RR rs) with s;
 [ replace (RR rt) with t | auto ]; auto.
constructor; auto.
(*MCRewriteIn rt0 '(R rt0)=(wi t).*)
match goal with
| H:(TrRegFile _ _ _ _),H5:(R rt = wi t) |- _ =>
    generalize (H rt (wi t) (RR rt) H5 (refl_equal (RR rt))); intros DDH
end.
inversion DDH; auto.
(*MCRewriteIn rs0 '(R rs0)=(wi s);*)
match goal with
| H:(TrRegFile _ _ _ _),H5:(R rs = wi s) |- _ =>
    generalize (H rs (wi s) (RR rs) H5 (refl_equal (RR rs))); intros DDH
end.
inversion DDH; auto.

unfold rfupd, updatereg in |- *; replace (beq_reg rd r) with false;
 try apply beq_neq_false; auto; replace (beq_regt r rd) with false;
 try apply beq_regt_false_neqR; try eauto;
 try (apply neq_reg_regt; auto).
Qed.



Lemma cpinv_preserv_movi :
 forall CH T rd i Is St (D : CpInv CH (basect T) St)
   (D1 : WFInstrSeq CH T (iseq (movi rd i) Is)) (D2 : curcmd St (_movi rd i)),
 CpInv CH (basect (rftupd T rd intty)) (Step St).
intros CH T rd i Is (MM, (RR, pc)); unfold curcmd, curmem, curpc in |- *;
 intros.
ElimEx6 D DLyt H R HT Ts G.
MCInv InstCodeTy.
MCRewriteIn T WFInstrSeq.
match goal with
| H0:(WFCHeap CH),H2:(WFHeap H HT),H1:(WFRegFile CH HT R G) |- _ =>
    generalize (progress_movi _ _ _ _ _ _ _ _ H0 H2 H1 D1);
     intros ((H'', (R'', Is'')), DDP');
     elim (preserv_movi _ _ _ _ _ _ _ _ _ _ _ H0 H2 H1 D1 DDP'); 
     intros HT'; IntrosSplitH
end.

unfold Step in |- *; replace (Dc (MM pc)) with (_movi rd i); auto.
Exists4 DLyt H'' R'' HT'.
Exists2 (nil (A:=Omega)) (rftupd G rd intty); split; auto.
constructor; auto.
split; auto.
split; auto.
split; auto.
split; auto.
MCInv Eval.
MCRewrite H''; unfold R' in |- *; MCRewrite R0; MCClear2 H'' R'; auto.
split; auto.

intros l hv DHH.
generalize (H5 _ _ DHH); intros (Ws, (a, DHH')); SplitAllHyps.
Exists2 Ws a; split; auto.
split; auto.
split; auto.
split; auto.
unfold updatereg in |- *.
replace (beq_regt rAlloc rd) with false; auto.
case rd; compute in |- *; auto.

split; auto.
(*Match Context With [H15:(?)=rd; H16:(?)=i|-?]
 -> (Unfold i0 in H15; Unfold i0 in H16).*)
MCRewrite rd0.

intro r; compare r rd; intro ee.
rewrite ee; unfold rfupd, updatereg in |- *;
 replace (beq_reg rd rd) with true; auto; replace (beq_regt rd rd) with true;
 auto; intros.
MCRewrite v; MCRewrite w; MCRewrite n; constructor; auto.

unfold rfupd, updatereg in |- *; replace (beq_reg rd r) with false;
 try apply beq_neq_false; auto; replace (beq_regt r rd) with false;
 try apply beq_regt_false_neqR; try eauto;
 try (apply neq_reg_regt; auto).
Qed.

Lemma cpinv_preserv_movl :
 forall CH T rd l w Is St hv T' (D : CpInv CH (basect T) St)
   (D1 : WFInstrSeq CH T (iseq (movl rd l) Is)) (D2 : curcmd St (_movi rd w))
   (D3 : fmaplook CH l hv) (D4 : typeofchv hv = T') 
   (D5 : Lytlook CLyt l w),
 CpInv CH (basect (rftupd T rd (codety T'))) (Step St).
intros CH T rd l w Is (MM, (RR, pc)); unfold curcmd, curmem, curpc in |- *;
 intros.
ElimEx6 D DLyt H R HT Ts G.
MCInv InstCodeTy.
MCRewriteIn T WFInstrSeq.
match goal with
| H0:(WFCHeap CH),H2:(WFHeap H HT),H1:(WFRegFile CH HT R G) |- _ =>
    generalize (progress_movl _ _ _ _ _ _ _ _ H0 H2 H1 D1);
     intros ((H'', (R'', Is'')), DDP');
     elim (preserv_movl _ _ _ _ _ _ _ _ _ _ _ H0 H2 H1 D1 DDP' _ _ D3 D4);
     intros HT'; IntrosSplitH
end.

unfold Step in |- *; replace (Dc (MM pc)) with (_movi rd w); auto.
Exists4 DLyt H'' R'' HT'.
Exists2 (nil (A:=Omega)) (rftupd G rd (codety T')); split; auto.
constructor; auto.
split; auto.
split; auto.
split; auto.
split; auto.
MCInv Eval.
MCRewrite H''; unfold R' in |- *; MCRewrite R0; MCClear2 H'' R'; auto.
split; auto.

intros l' hv' DHH.
generalize (H5 _ _ DHH); intros (Ws, (a, DHH')); SplitAllHyps.
Exists2 Ws a; split; auto.
split; auto.
split; auto.
split; auto.
unfold updatereg in |- *.
replace (beq_regt rAlloc rd) with false; auto.
case rd; compute in |- *; auto.

split; auto.
(*Match Context With [H15:(?)=rd; H16:(?)=l|-?]
 -> (Unfold i in H15; Unfold i in H16).*)
MCRewrite rd0.

intro r; compare r rd; intro ee.
rewrite ee; unfold rfupd, updatereg in |- *;
 replace (beq_reg rd rd) with true; auto; replace (beq_regt rd rd) with true;
 auto; intros.
MCRewrite v; MCRewrite w0; MCRewrite l0.
constructor; auto.

unfold rfupd, updatereg in |- *; replace (beq_reg rd r) with false;
 try apply beq_neq_false; auto; replace (beq_regt r rd) with false;
 try apply beq_regt_false_neqR; try eauto;
 try (apply neq_reg_regt; auto).
Qed.





Lemma oeq_wfwordval :
 forall CH HT w t t', oeq t t' -> WFWordVal CH HT w t -> WFWordVal CH HT w t'.
intros CH HT w t t' H; MCInv oeq.
intros.
MCInv WFWordVal.
apply label_wval0 with G Is; auto.
apply trans_octeq with F; auto.

apply label_wval1 with G; auto.
apply trans_octeq with F; auto.

elim H5; eauto.
Qed.


Lemma octeq_wfregfile :
 forall CH HT R G G',
 WFRegFile CH HT R G -> octeq (basect G) (basect G') -> WFRegFile CH HT R G'.
unfold WFRegFile, flook in |- *; intros.
MCInv octeq; auto.
generalize (H r (G r) (refl_equal _)); intros.
generalize (H4 r); intros.
apply oeq_wfwordval with (G r); auto.
MCRewrite t.
Qed.



Lemma XTALtoCAPcmds :
 forall DLyt CT CH Is T Ws Cs (D0 : IsCpGen CT CH)
   (D1 : ExtLytPred CLyt DLyt CH) (D2 : TrCHeapVal CLyt (code T Is) Ws)
   (D3 : Cs = map Dc Ws)
   (D4 : forall Ts G, InstCodeTy T Ts G -> WFInstrSeq CH G Is),
 WFCapCmds TrivSP CT (CpInv CH T) Cs.
induction Is; intros.
induction i; intros; MCInv TrCHeapVal; MCClear2 T0 Ws0;
 MCRewriteIn (map Dc Ws) TrInstrSeq; MCInv TrInstrSeq;
  (* MCClear2 Ts Ns; *)
 MCClear4 i Is1 Is0 cs; MCInv TrInstr; MCRewriteIn Cs TrInstr;
 MCRewriteIn Cs TrInstrSeq.
MCClear4 Cs1 rd rs rt; MCRewriteIn Cs (Cs = map Dc Ws); MCClear Cs.

apply wfcapcmd with (reg2regt r) (reg2regt r0) (reg2regt r1) 0.
tauto.
intros; compute in |- *; auto.
intros (MM, (RR, pc)); intros.
MCElimEx4 Lyt' H' R' HT' (CpInv CH T (MM, (RR, pc))).
MCElimEx2 Ts' G' Omegalist.
assert (CpInv CH (basect G') (MM, (RR, pc))).
  Exists4 Lyt' H' R' HT'; Exists2 (nil (A:=Omega)) G'; split;
   [ constructor | split; [ auto | split; auto ] ].
exists (CpInv CH (basect (rftupd G' r intty))); split; auto.

apply cpinv_preserv_add with r0 r1 Is; auto.
apply D4 with Ts'; auto.

assert (exists w', (exists Ws', Ws = w' :: Ws' /\ Cs0 = map Dc Ws')).
   induction  Ws as [| a Ws HrecWs]; eauto.
   unfold map in D3; discriminate D3.
   Exists2 a Ws; unfold map in D3; injection D3; auto.
MCElimEx2 w' Ws' Word.
apply IHIs with Ws'; auto.
constructor.
MCRewriteIn Cs0 TrInstrSeq.

intros Ts'' G'' HI.
inversion HI.
match goal with
| H1:(InstCodeTy T Ts' G') |- _ => generalize (D4 _ _ H1); intros HHI
end.
inversion HHI; auto.
(* DONE ADD! *)

(* MOVI *)
apply wfcapcmd with (reg2regt r) (reg2regt r) (reg2regt r) n.
tauto.
intros; compute in |- *; auto.
intros (MM, (RR, pc)); intros.
MCElimEx4 Lyt' H' R' HT' (CpInv CH T (MM, (RR, pc))).
MCElimEx2 Ts' G' Omegalist.
assert (CpInv CH (basect G') (MM, (RR, pc))).
  Exists4 Lyt' H' R' HT'; Exists2 (nil (A:=Omega)) G'; split;
   [ constructor | split; [ auto | split; auto ] ].
exists (CpInv CH (basect (rftupd G' r intty))); split; auto.

apply cpinv_preserv_movi with n Is; auto.
apply D4 with Ts'; auto.

assert (exists w', (exists Ws', Ws = w' :: Ws' /\ Cs0 = map Dc Ws')).
   induction  Ws as [| a Ws HrecWs]; eauto.
   rewrite <- H0 in D3; unfold map in D3; discriminate D3.
   Exists2 a Ws; rewrite <- H0 in D3; unfold map in D3; injection D3; auto.
MCElimEx2 w' Ws' Word.
apply IHIs with Ws'; auto.
constructor.
MCRewriteIn Cs0 TrInstrSeq.

intros Ts'' G'' HI.
inversion HI.
match goal with
| H1:(InstCodeTy T Ts' G') |- _ => generalize (D4 _ _ H1); intros HHI
end.
inversion HHI; auto.
(* DONE MOVI *)

(* MOVL *)
apply wfcapcmd with (reg2regt r) (reg2regt r) (reg2regt r) w.
tauto.
intros; compute in |- *; auto.
intros (MM, (RR, pc)); intros.
MCElimEx4 Lyt' H' R' HT' (CpInv CH T (MM, (RR, pc))).
MCElimEx2 Ts' G' Omegalist.
assert (CpInv CH (basect G') (MM, (RR, pc))).
  Exists4 Lyt' H' R' HT'; Exists2 (nil (A:=Omega)) G'; split;
   [ constructor | split; [ auto | split; auto ] ].

generalize (D4 _ _ H9); intros WFI; inversion WFI.
exists (CpInv CH (basect (rftupd G' r (codety T0)))); split; auto.

apply cpinv_preserv_movl with n w Is hv; auto.

assert (exists w', (exists Ws', Ws = w' :: Ws' /\ Cs0 = map Dc Ws')).
   induction  Ws as [| a Ws HrecWs]; eauto.
   rewrite <- H1 in D3; unfold map in D3; discriminate D3.
   Exists2 a Ws; rewrite <- H1 in D3; unfold map in D3; injection D3; auto.
MCElimEx2 w' Ws' Word.
apply IHIs with Ws'; auto.
constructor.
MCRewriteIn Cs0 TrInstrSeq.

intros Ts'' G'' HI.
inversion HI.
match goal with
| H1:(InstCodeTy T Ts' G') |- _ => generalize (D4 _ _ H1); intros HHI
end.
inversion HHI; auto.
(* DONE MOVL *)

(* JD *)
MCInv TrCHeapVal; MCClear2 T0 Ws0.
MCRewriteIn (map Dc Ws) TrInstrSeq; MCInv TrInstrSeq. (*MCClear4 l Ts Ns Is.*)
apply wfcapjd; intros (MM, (RR, pc)); intros;
 try match goal with
     |  |- (TrivSP _) => compute in |- *; auto
     end.

match goal with
| H:(CpInv _ _ _) |- _ => ElimEx H DLyt'; rename H into CPH
end.
MCClear3 Is l Ts.
ElimEx5 CPH H R HT Ts G. (* Ts Ns G '(list Omega).*)

match goal with
| H1:(InstCodeTy T Ts G) |- _ => generalize (D4 _ _ H1); intros HHI
end.
MCInv (WFInstrSeq CH G (jd n o)).
MCRewriteIn T0 InstCodeTy; MCClear5 Ts0 T0 l Ch G0.
assert (WFCHeapVal CH hv).
   MCUnfoldInSelf WFCHeap; MCUnfoldInSelf ExtLytPred; SplitAllHyps.
   match goal with
   | H6:(eqdomf _ _ _ _ _),H4:(fmaplook CH n _),H8:(
   forall l chv, _ -> WFCHeapVal CH chv) |- _ => apply H8 with n; auto
   end.

MCInv WFCHeapVal.

(*** CASE WHERE JUMPING TO XTAL CODE ***)
generalize D0; intros (_, (DD2, _)).
(* Cut (ExtCode w)==(noneT ?); [ Intros DD3 | Auto ]. *)
MCRewriteIn hv (fmaplook CH n hv).
MCRewriteIn hv (InstCodeTy (typeofchv hv) o G').
MCUnfoldIn typeofchv InstCodeTy.

assert (Lytlook CLyt n w).
  unfold Lytlook, fmaplook in |- *; repeat MCUnfoldInSelf Lytlook;
   UnfoldFmaps.

match goal with
| H4:(fmaplook CH n _),H17:(Lytlook CLyt n _) |- _ =>
    generalize (DD2 _ _ _ _ H4 H17); intros
end.
exists (InstrSeq_size Is); exists (CpInterf CH T0); split; auto.

match goal with
| H1:(curcmd (MM, (RR, pc)) (_jd w)) |- _ =>
    unfold curcmd, curmem, curpc in H1; unfold Step in |- *; rewrite H1
end.
exists DLyt'; Exists3 H R HT.
Exists2 o G'; split; auto.
split; auto.
split; auto.
split; auto.
apply octeq_wfregfile with G; auto.


(*** CASE WHERE JUMPING TO EXTERNAL CODE! ***)
MCRewriteIn hv (fmaplook CH n hv).
MCRewriteIn hv (InstCodeTy (typeofchv hv) o G').
MCUnfoldIn typeofchv InstCodeTy.
generalize D1; intros (D2on, _).
match goal with
| H15:(fmaplook CH n (stub _)),H10:(Lytlook CLyt n w) |- _ =>
    generalize (cdspec_heapstub_extcode CT CH n T0 w D2on D0 H15 H10)
end.
intros (Cs', (P', DDH)); SplitAllHyps.
exists (length Cs'); exists P'; split; auto.
match goal with
| H18:(forall St, CpInterf _ _ _ -> P' St) |- _ => 
apply H18
end.

unfold Step in |- *; replace (Dc (MM pc)) with (_jd w).
Exists4 DLyt' H R HT.
Exists2 o G'; split; auto.
split; auto.
split; auto.
split; auto.
apply octeq_wfregfile with G; auto.

(* END OF JD! *)


(* JMP *)
MCInv TrCHeapVal; MCClear2 T0 Ws0; MCRewriteIn (map Dc Ws) TrInstrSeq;
 MCInv TrInstrSeq.
MCClear2 Ts Is.
apply wfcapjmp; intros (MM, (RR, pc)); intros;
 try match goal with
     |  |- (TrivSP _) => compute in |- *; auto
     end.

match goal with
| H:(CpInv _ _ _) |- _ => ElimEx H DLyt'; rename H into CPH
end.
ElimEx5 CPH H R HT Ts G. 

match goal with
| H1:(InstCodeTy T Ts G) |- _ => generalize (D4 _ _ H1); intros HHI
end.
MCInv (WFInstrSeq CH G (jmp r o)).

generalize (H7 (*WFRegFile*)  _ _ H14); intros WFR.
assert
 (exists hv,
    (exists l,
       octeq (typeofchv hv) T0 /\
       WFCHeapVal CH hv /\ fmaplook CH l hv /\ R r = wl l)).
  MCUnfoldInSelf WFCHeap; MCUnfoldInSelf ExtLytPred; SplitAllHyps.
  inversion WFR.
  Exists2 (code G1 Is) l; split; auto. split; auto.
  apply H22 with l; auto.
  Exists2 (stub G1) l; split; auto. split; auto.
  apply H22 with l; auto.
  elim H26; eauto.
MCElimEx2 hv l CHeapVal.
assert (Lytlook CLyt l (RR r)).
  unfold Lytlook, fmaplook in |- *; repeat MCUnfoldInSelf Lytlook;
   UnfoldFmaps.
  generalize (H10 (*TrRegFile*)  _ _ _ H23 (refl_equal (RR r))); intros HTR.
  inversion HTR; auto.
  MCInv ExtLytPred; SplitAllHyps.
  generalize (H26 (*eqdomf*)  l); unfold iff in |- *; IntrosSplitH.
  assert (indomf CLyt l).
    apply H29; unfold indomf, fmaplook in |- *; eauto.
  MCUnfoldInSelf disjf.
  repeat MCUnfoldInSelf Lytlook; UnfoldFmaps.
  elim (H27 l); split; unfold indomf, fmaplook in |- *; eauto.
MCInv WFCHeapVal.

(*** CASE WHERE JUMPING TO XTAL CODE ***)
generalize D0; intros (_, (DD2, _)).
MCRewriteIn hv (fmaplook CH l hv).
  (*MCRewriteIn T0 InstCodeTy.
    MCRewriteIn hv '(InstCodeTy (typeofchv hv) o G).
    MCUnfoldIn typeofchv InstCodeTy.*)
MCRewriteIn hv octeq; MCUnfoldIn typeofchv octeq.
match goal with
| H4:(fmaplook CH l _),H17:(Lytlook CLyt l _) |- _ =>
    generalize (DD2 _ _ _ _ H4 H17); intros
end.
exists (InstrSeq_size Is); exists (CpInterf CH T1); split; auto.

  (*Unfold currf; Replace T0 with T1; Auto.
    MCRewriteIn hv '(typeofchv hv)=T0.*)

unfold curcmd, curmem, curpc in H1; unfold Step in |- *; rewrite H1.
exists DLyt'; Exists3 H R HT.

elim (octeq_instcodety _ _ _ _ (sym_octeq _ _ H20) H17); intros G'';
 IntrosSplitH.
Exists2 o G''.
split; auto.
split; auto.
split; auto.
split; auto.
apply octeq_wfregfile with G; auto.
apply trans_octeq with (basect G'); auto.

(*** CASE WHERE JUMPING TO EXTERNAL CODE! ***)
MCRewriteIn hv (fmaplook CH l hv).
   (*MCRewriteIn T0 InstCodeTy.
     MCRewriteIn hv '(InstCodeTy (typeofchv hv) o G).
     MCUnfoldIn typeofchv InstCodeTy.*)
MCRewriteIn hv octeq; MCUnfoldIn typeofchv octeq.

generalize D1; intros (D2on, _).

match goal with
| H15:(fmaplook CH l (stub _)),H10:(Lytlook CLyt l _) |- _ =>
    generalize (cdspec_heapstub_extcode CT CH l T1 _ D2on D0 H15 H10)
end.
intros (Cs', (P', DDH)); SplitAllHyps.
exists (length Cs'); exists P'; split; auto.
match goal with
| H18:(forall St, CpInterf _ _ _ -> P' St) |- _ => 
apply H18
end.

unfold Step in |- *; replace (Dc (MM pc)) with (_jmp r).
Exists4 DLyt' H R HT.

elim (octeq_instcodety _ _ _ _ (sym_octeq _ _ H20) H17); intros G'';
 IntrosSplitH.
Exists2 o G''; split; auto.
split; auto.
split; auto.
split; auto.
apply octeq_wfregfile with G; auto.
apply trans_octeq with (basect G'); auto.
(* END OF JMP! *)
Qed.


Lemma XTALtoCAPcdspec :
 forall DLyt CT CH H R Is MM RR pc (D0 : IsCpGen CT CH)
   (D1 : WFProgram (CH, (H, (R, Is))))
   (D2 : TrProgram ExtCode CLyt DLyt (CH, (H, (R, Is))) (MM, (RR, pc)))
   (extcode_wf : forall f Cs P,
                 ExtCode f = someT (Cs,* P) ->
                 forall CT', IsCpGen CT' CH -> WFCapCmds TrivSP CT' P Cs),
 WFCapCdSpec TrivSP MM CT.
intros; intro; intros.
generalize D0; intros (_, (_, (_, (D0', _)))).
elim (D0' f n P); clear D0'; try intros D0'; auto.

MCElimEx Cs (list Cmd).
exists Cs; split; [ auto | split; auto ].
MCInv TrProgram.
apply cmdsin_flatten.
match goal with
| H15:(forall w Cs Pcs, ExtCode w = _ -> _) |- _ => 
apply H15 with P; auto
end.
apply extcode_wf with f; auto.


MCElimEx3 l' G' Is' int.
MCInv TrProgram; MCClear6 MM0 RR0 pc0 Is0 CH0 R0.
cut (TrCHeap CLyt CH MM); [ intros DH | auto ].
match goal with
| H1:(fmaplook CH l' (code G' Is')) |- _ =>
    generalize (DH l' (code G' Is') H1); clear DH; intros (Ws, (a, DH));
     SplitAllHyps
end.
exists (map Dc Ws); split; auto.
apply cmdsin_flatten; exists Ws; split; auto.
try replace f with a; auto.
repeat MCUnfoldInSelf Lytlook; UnfoldFmaps.
MCInjection (CLyt l').
split; auto.

MCInv TrCHeapVal.
MCRewrite n; apply TrInstrSeq_size_eq with CLyt; auto.

MCRewrite P.
apply XTALtoCAPcmds with DLyt Is' Ws; auto.

inversion D1;
 match goal with
 | H22:(WFCHeap CH),H3:(fmaplook CH l' (code G' Is')) |- _ =>
     generalize H22; intros (DLIM, DWFI); generalize (DWFI _ _ H3);
      intros DWF; MCInv WFCHeapVal
 end.
Qed.




(*************************************************************************)
(* MAIN THEOREM *)
Theorem XTALtoCAP :
 forall CT DLyt CH H R Is St,
 IsCpGen CT CH ->
 WFProgram (CH, (H, (R, Is))) ->
 TrProgram ExtCode CLyt DLyt (CH, (H, (R, Is))) St ->
 forall
   extcode_wf : forall f Cs P,
                ExtCode f = someT (Cs,* P) ->
                forall CT', IsCpGen CT' CH -> WFCapCmds TrivSP CT' P Cs,
 WFCapState TrivSP St.
intros; MCInv WFProgram; MCClearAll.
MCInv TrProgram; MCRewriteIn St TrProgram; MCClear St; MCClearAll.
assert
 (exists Cs,
    (exists Ws,
       TrInstrSeq CLyt Is Cs /\ WordsInMem Ws pc MM /\ Cs = map Dc Ws)).
  MCInv TrPC.
  MCElimEx T CodeTy; MCElimEx Is' InstrSeq; MCElimEx wpc Word;
   MCElimEx nis int.
  match goal with
  | H1:(TrCHeap _ CH _),H2:(fmaplook CH _ _) |- _ =>
      generalize (H1 _ _ H2); intros (Ws', (apc, DD)); SplitAllHyps
  end.
  
  MCInv TrCHeapVal; MCClear3 Is0 T0 Ws.
  assert
   (exists Ws : WordList,
      tailnth nis _ Ws' = Some Ws /\
      TrInstrSeq CLyt Is (map Dc Ws) /\ WordsInMem Ws (nis + wpc) MM).
     apply TrInstrSeq_subseq_tailnth with Is'; auto.
     replace wpc with apc; repeat MCUnfoldInSelf Lytlook; UnfoldFmaps;
      MCInjection (CLyt x); auto.
  MCElimEx Ws WordList; Exists2 (map Dc Ws) Ws.
  split; [ auto | split; auto ].
  MCRewrite pc; replace (wpc + nis) with (nis + wpc); auto.
  ring.
MCElimEx Cs (list Cmd); MCElimEx Ws (list Word).

apply wfcapstate with CT Cs (CpInterf CH (basect G)); auto.

(* CAP code memory well-formedness *)
apply XTALtoCAPcdspec with DLyt CH H R Is RR pc; auto.

(* Flatten *)
apply cmdsin_flatten.
  MCRewrite Cs; unfold CmdsInMem in |- *; exists Ws; auto.

(* commands of the current pc is well-formed *)
apply XTALtoCAPcmds with DLyt Is Ws; auto.
constructor; MCRewriteRev Cs.
intros Ts' G' HI; inversion HI; MCRewrite G'; auto.

(* CpInv holds trivially on top *)
Exists4 DLyt H R HT; Exists2 (nil (A:=Omega)) G; split;
 [ constructor | split; auto ].
split; auto.

(* pc is in code area? *)
apply IsCodeAreaTrPC with CH DLyt MM Is Ws; auto.
Qed.



End Proofs.










(**************************************************************************)
(**************************************************************************)
(**************************************************************************)
(**************************************************************************)
(* External Array Library *)

Definition lab_arroutbounds := 0.
Definition lab_arrayget := 1.


Definition cmds_arrayget : list Cmd :=
  _ld rI rA 0
  :: _addi rB rB 1
     :: _bgt rB rI lab_arroutbounds
        :: _add rA rA rB :: _ld rA rA 0 :: _jmp rH :: nil.

Definition type_arrayget : CodeTy :=
  tabsct
    (tabsctl _
       (tabsctl _
          (tabsctl _
             (tabsctl _
                (basectl _
                   (fun r =>
                    match r with
                    | r0 => liftLzeroto 5 arrtyL
                    | r1 => liftLzeroto 5 inttyL
                    | r2 => varL 4
                    | r3 => liftL _ (varL 3)
                    | r4 => liftL _ (liftL _ (varL 2))
                    | r5 => liftL _ (liftL _ (liftL _ (varL 1)))
                    | r6 => liftL _ (liftL _ (liftL _ (liftL _ (varL 0))))
                    | r7 =>
                        codetyL _
                          (tabsctl _
                             (tabsctl _
                                (basectl _
                                   (fun r' =>
                                    match r' with
                                    | r0 => liftLzeroto 7 inttyL
                                    | r1 =>
                                        liftL _
                                          (liftL _
                                             (liftL _
                                                (liftL _ (liftL _ (varL 1)))))
                                    | r2 => varL 6
                                    | r3 => liftL _ (varL 5)
                                    | r4 => liftL _ (liftL _ (varL 4))
                                    | r5 =>
                                        liftL _ (liftL _ (liftL _ (varL 3)))
                                    | r6 =>
                                        liftL _
                                          (liftL _
                                             (liftL _ (liftL _ (varL 2))))
                                    | r7 =>
                                        liftL _
                                          (liftL _
                                             (liftL _
                                                (liftL _ (liftL _ (varL 1)))))
                                    end))))
                    end)))))).

(* see the better (more modular) predicate for arrayget following this bunch
   of definitions and associated lemmas *)

Definition pred_arrayget clyt ch : Pred := CpInv clyt ch type_arrayget.

Definition ExtCode CLyt CH : ExtCodeTy :=
  fun n =>
  match n with
  | 0 => someT (_jd 0 :: nil,* TrivSP)
  | 1 => someT (cmds_arrayget,* pred_arrayget CLyt CH)
  | _ => noneT _
  end.

Require Import natutil.

Lemma extcodewf_arrayget :
 forall CLyt CH CT (D1 : IsCpGen (ExtCode CLyt CH) CLyt CT CH),
 WFCapCmds TrivSP CT (pred_arrayget CLyt CH) cmds_arrayget.
assert True; try intros H; auto.
unfold cmds_arrayget, pred_arrayget in |- *; intros.
apply wfcapcmd with rI rA rA 0; try tauto; try intros (MM, (RR, pc)); intros.
exists (fun St => St = (MM, (updatereg RR rI (MM (RR rA)), S pc))).
split; auto.
unfold curcmd, curmem, curpc in H1; unfold Step in |- *; rewrite H1.
replace (RR rA + 0) with (RR rA); auto.

apply wfcapcmd with rB rB rA 1; try tauto; try intros (MM1, (RR1, pc1)).
unfold curcmd, currf, curmem, curpc in |- *; intros.
unfold Step in |- *; rewrite H3; auto.
exists (fun St => St = (MM1, (updatereg RR1 rB (RR1 rB + 1), S pc1))); split;
 auto.

constructor.
intros; compute in |- *; auto.
unfold curcmd, currf, curmem, curpc in |- *; intros (MM2, (RR2, pc2)); intros.
Exists2 1 TrivSP; split; compute in |- *; auto.
MCInv IsCpGen; SplitAllHyps.
replace 1 with (length (_jd 0 :: nil)); auto.

unfold curcmd, currf, curmem, curpc in |- *; intros (MM2, (RR2, pc2)); intros.
unfold Step in |- *; rewrite H5; auto.
exists (fun St => St = (MM2, (RR2, S pc2))); split; auto.
replace (blt_nat (RR2 rI) (RR2 rB)) with false; auto.
symmetry  in |- *; apply le_blt_false; auto.

apply wfcapcmd with rA rA rB 0; try tauto; try intros (MM3, (RR3, pc3)).
unfold curcmd, currf, curmem, curpc in |- *; intros.
unfold Step in |- *; rewrite H8; auto.
exists (fun St => St = (MM3, (updatereg RR3 rA (RR3 rA + RR3 rB), S pc3)));
 split; auto.

apply wfcapcmd with rA rA rA 0; try tauto; try intros (MM4, (RR4, pc4)).
unfold curcmd, currf, curmem, curpc in |- *; intros.
unfold Step in |- *; rewrite H10; auto.
exists (fun St => St = (MM4, (updatereg RR4 rA (MM4 (RR4 rA + 0)), S pc4)));
 split; auto.

constructor.
intros; compute in |- *; auto.
unfold curcmd, currf, curmem, curpc in |- *; intros (MM5, (RR5, pc5)); intros.
unfold Step in |- *; rewrite H12; auto.
clear H. ElimEx6 H0 DLyt H R HT Ts G.
MCInvClear InstCodeTy; MCClear3 F Ts G0.
do 4
 (MCInvClear ListSubstL; ElimAllExistS; MCRewriteIn F ListSubstL; MCClear F;
   match goal with
   | H21:(_ = G0),H23:(_ = G) |- _ =>
       rewrite <- H21 (*..=G0*)  in H23 (*..=G*); MCClear G0;
        rename G1 into G0
   end).
MCInvClear ListSubstL; ElimAllExistS.
match goal with
| H23:(_ = G),H32:(_ = G0),H30:(F = _) |- _ =>
    rewrite H30 in H32; MCClear F; rewrite <- H32 in H23; MCClear G0
end.

assert (WFWordVal CH HT (R r7) (G r7)).
  apply H15 (*WFRegFile*); compute in |- *; auto.
MCInv WFWordVal; MCClear2 HT0 Ch.
MCRewriteIn G (intty = G r7); inversion H36 (*intty=...*).

(* JUMPING BACK TO TAL CODE *)
unfold IsCpGen in D1; SplitAllHyps.
generalize (H16 (*TrCHeap*)  _ _ H36 (*fmaplook CH*)); intros;
 MCElimEx2 Ws a WordList.
generalize (H38 _ _ _ _ H36 H42); intros.
Exists2 (InstrSeq_size Is) (CpInv CLyt CH G0); split; auto.
replace (RR5 rH) with a; auto.
replace (RR5 rH) with (RR rH).
assert (TrWordVal CLyt DLyt (wl l) (RR rH)).
  unfold TrRegFile in H18; apply H18 with r7 (*TrRegFile*); auto.
MCInv TrWordVal; repeat MCUnfoldInSelf Lytlook; UnfoldFmaps;
 MCInjection (CLyt l).
MCInv ExtLytPred; SplitAllHyps.
generalize (H50 (*disjf*)  l); intros BAD; elim BAD;
 unfold indomf, fmaplook in |- *; split; eauto.

repeat match goal with 
  | H:(_, (?h, _))=_ |- _ => (injection H; intros; MCRewrite h) end.


(* really we should instantiate with updated R, but anyway, just
   to check the reasoning we will instantiate with the same R
   and use (apply Super) to fudge around the details :) *)
Exists4 DLyt H R HT. 
exists (intty :: G r7 :: nil).

MCRewriteIn G (codety G' = G r7); generalize H33.
unfold SubstL at 5, subst_aux in |- *; unfold SubstL at 4, subst_aux in |- *;
 unfold SubstL at 3, subst_aux in |- *; unfold SubstL at 2, subst_aux in |- *;
 unfold SubstL at 1, subst_aux in |- *;
 fold subst_aux subst_codety_aux.
unfold RUnlift at 1 in |- *; fold RUnliftCodeTy in |- *.
EQREC
 (eq_rec 0 CodeTyL
    (tabsctl 0
       (tabsctl 1
          (basectl 2
             (fun r : Reg =>
              subst_aux 3
                (subst_aux 4
                   (subst_aux 5
                      (subst_aux 6
                         (subst_aux 7
                            match r with
                            | r0 => liftLzeroto 7 inttyL
                            | r1 =>
                                liftL 6
                                  (liftL 5
                                     (liftL 4 (liftL 3 (liftL 2 (varL 1)))))
                            | r2 => varL 6
                            | r3 => liftL 6 (varL 5)
                            | r4 => liftL 6 (liftL 5 (varL 4))
                            | r5 => liftL 6 (liftL 5 (liftL 4 (varL 3)))
                            | r6 =>
                                liftL 6
                                  (liftL 5 (liftL 4 (liftL 3 (varL 2))))
                            | r7 =>
                                liftL 6
                                  (liftL 5
                                     (liftL 4 (liftL 3 (liftL 2 (varL 1)))))
                            end 6 (f_equal S (f_equal S (refl_equal 5)))
                            (liftL 5 (liftL 4 (liftLzeroto 4 (RLift t3))))) 5
                         (f_equal S (f_equal S (refl_equal 4)))
                         (liftL 4 (liftL 3 (liftLzeroto 3 (RLift t2))))) 4
                      (f_equal S (f_equal S (refl_equal 3)))
                      (liftL 3 (liftL 2 (liftLzeroto 2 (RLift t1))))) 3
                   (f_equal S (f_equal S (refl_equal 2)))
                   (liftL 2 (liftL 1 (liftLzeroto 1 (RLift t0))))) 2
                (f_equal S (f_equal S (refl_equal 1)))
                (liftL 1 (liftL 0 (liftLzeroto 0 (RLift t)))))))) 0 OO)
 (tabsctl 0
    (tabsctl 1
       (basectl 2
          (fun r : Reg =>
           subst_aux 3
             (subst_aux 4
                (subst_aux 5
                   (subst_aux 6
                      (subst_aux 7
                         match r with
                         | r0 => liftLzeroto 7 inttyL
                         | r1 =>
                             liftL 6
                               (liftL 5
                                  (liftL 4 (liftL 3 (liftL 2 (varL 1)))))
                         | r2 => varL 6
                         | r3 => liftL 6 (varL 5)
                         | r4 => liftL 6 (liftL 5 (varL 4))
                         | r5 => liftL 6 (liftL 5 (liftL 4 (varL 3)))
                         | r6 =>
                             liftL 6 (liftL 5 (liftL 4 (liftL 3 (varL 2))))
                         | r7 =>
                             liftL 6
                               (liftL 5
                                  (liftL 4 (liftL 3 (liftL 2 (varL 1)))))
                         end 6 (f_equal S (f_equal S (refl_equal 5)))
                         (liftL 5 (liftL 4 (liftLzeroto 4 (RLift t3))))) 5
                      (f_equal S (f_equal S (refl_equal 4)))
                      (liftL 4 (liftL 3 (liftLzeroto 3 (RLift t2))))) 4
                   (f_equal S (f_equal S (refl_equal 3)))
                   (liftL 3 (liftL 2 (liftLzeroto 2 (RLift t1))))) 3
                (f_equal S (f_equal S (refl_equal 2)))
                (liftL 2 (liftL 1 (liftLzeroto 1 (RLift t0))))) 2
             (f_equal S (f_equal S (refl_equal 1)))
             (liftL 1 (liftL 0 (liftLzeroto 0 (RLift t)))))))).
intros HEQ; injection HEQ; clear HEQ; intros HEQ.

MCRewriteIn G' octeq; MCInv octeq.
MCInv octleq; ElimAllExistS; MCRewrite F.
MCRewriteIn F'0 octleq; MCInv octleq; ElimAllExistS; MCRewrite F0.
exists
 (fun r =>
  RUnlift 0
    ((fun r =>
      SubstL 0 ((fun r => SubstL 1 (F1 r) (liftLzeroto _ (RLift (G r7)))) r)
        (liftLzeroto _ (RLift intty))) r) OO).
split; auto.
assert
 (ListSubstL 0 (tabsctl 1 (basectl 2 F1)) (intty :: G r7 :: nil)
    (fun r : Reg =>
     SubstL 0 (SubstL 1 (F1 r) (liftLzeroto 1 (RLift (G r7))))
       (liftLzeroto 0 (RLift intty)))).
  assert
   (ListSubstL 1 (basectl 2 F1) (G r7 :: nil)
      (fun r : Reg => SubstL 1 (F1 r) (liftLzeroto 1 (RLift (G r7))))).
    constructor.
  generalize
   (lsub_tabs _ (basectl 2 F1) intty (G r7 :: nil)
      (fun r => SubstL 1 (F1 r) (liftLzeroto _ (RLift (G r7)))) H56); 
   intros.
  auto.
generalize
 (instct_tabs (tabsctl 1 (basectl 2 F1)) (intty :: G r7 :: nil)
    (fun r =>
     SubstL 0 ((fun r => SubstL 1 (F1 r) (liftLzeroto _ (RLift (G r7)))) r)
       (liftLzeroto _ (RLift intty))) H56).
intros; auto.
split; auto.
split; auto.
split; auto.

apply octeq_wfregfile with G; auto.
constructor; intros.
match goal with 
  | H54:(F'1=_),H55:(forall r:Reg, oleq _ _ (F'1 r)) |- _ 
      => rewrite H54 in H55; generalize (H55 r); induction  r as [| | | | | | | ];
         intros
end.
assert (oleq 2 (F1 r0) (liftL _ (liftL _ inttyL))).
  compute in H56; auto.
replace (F1 r0) with (liftL 1 (liftL 0 inttyL)).
MCRewrite G; compute in |- *.
apply Super. 
(* OKAY IT_S WRONG *)

inversion H57; ElimAllExistS.
rewrite <- H60; rewrite H59; auto.
rewrite <- H59; rewrite H60 in H61; inversion H61; ElimAllExistS.
rewrite <- H64; rewrite H63; auto.
rewrite <- H63; rewrite H64 in H65; inversion H65; ElimAllExistS.
rewrite <- H68; rewrite H67; auto.

apply Super.
apply Super.
apply Super.
apply Super.
apply Super.
apply Super.
apply Super.

replace MM5 with MM.
split; auto.
split; auto.
apply Super. (* invariant on RR5 *)
split; auto.
apply Super. (* well this is not quite *)

repeat match goal with 
  | H:(?h, _)=_ |- _=?h => (injection H; intros; MCRewrite h) end.
(* DONE WITH JUMPING TO XTAL CODE *)

(* JUMPING TO ANOTHER STUB *)
apply Super.

(* DONE WITH JUMPING TO ANOTHER STUB *)

MCRewriteIn G not; elim H36 (*intty=...*); compute in |- *; eauto.
Qed.






Lemma extcodewf_arraylib
  : forall CLyt CH f,
    let EC := ExtCode CLyt CH in
    forall Cs P,
    EC f = someT (Cs,*P) 
     -> forall CT, IsCpGen EC CLyt CT CH -> WFCapCmds TrivSP CT P Cs.
intros CLyt CH f EC; unfold EC, ExtCode at 1.
induction f.

intros Cs P dinj; injection dinj; intros dinj1 dinj2.
MCRewrite2 Cs P; clear dinj dinj1 dinj2; clear Cs P.
intros.
constructor; intros; auto.
Exists2 1 TrivSP; split; auto.
MCUnfoldInSelf IsCpGen; SplitAllHyps.
assert (ExtCode CLyt CH 0 = someT (_jd 0 :: nil,* TrivSP)).
  compute; auto.
match goal with 
 | H:ExtCode _ _ _ = _ |- _ => generalize (H2 _ _ _ H) end;
compute; auto.

induction f.
intros Cs P dinj; injection dinj; intros dinj1 dinj2.
MCRewrite2 Cs P; clear dinj dinj1 dinj2; clear Cs P.
intros.
apply extcodewf_arrayget; auto.

intros Cs P disc; discriminate disc.
Save.


(* alternate precondition style for array_get *)

Definition pred_alt_arrayget (Q:Pred) : Pred :=
  fun St => Q(St).

Definition ExtCode_alt Q_arrayget : ExtCodeTy :=
  fun n =>
  match n with
  | 0 => someT (_jd 0 :: nil,* TrivSP)
  | 1 => someT (cmds_arrayget,* pred_alt_arrayget Q_arrayget)
  | _ => noneT _
  end.

Lemma extcodewf_arrayget_alt : 
 forall (CT:CdSpec) (Q_arrayget:Pred)
   (DC1:(CT 0)=someT(1,*TrivSP)),
   (forall St, match St with (MM,(RR,pc)) =>
     (Q_arrayget St) -> 
     (exists n,exists Q, CT(RR r7)=someT(n,*Q) /\
       forall RR':RFile, 
	           (RR' r0)=(MM (S (RR r0)+(RR r1))) /\
	           (RR' r2)=(RR r2) /\
		   (RR' r3)=(RR r3) /\
		   (RR' r4)=(RR r4) /\
		   (RR' r5)=(RR r5) /\
		   (RR' r6)=(RR r6) ->
		   Q(MM,(RR',RR r7))) end) ->
   WFCapCmds TrivSP CT (pred_alt_arrayget Q_arrayget) cmds_arrayget.
Proof.
intros.
unfold pred_alt_arrayget, cmds_arrayget.
apply wfcapcmd with rI rA rA 0; try unfold TrivSP; 
  try tauto; try intros (MM, (RR, pc)); intros.
exists (fun St => St = (MM, (updatereg RR rI (MM (RR rA)), S pc))).
split; auto.
unfold curcmd, curmem, curpc in H1; unfold Step in |- *; rewrite H1.
replace (RR rA + 0) with (RR rA); auto.

apply wfcapcmd with rB rB rA 1; try tauto; try intros (MM1, (RR1, pc1)).
unfold curcmd, currf, curmem, curpc in |- *; intros.
unfold Step in |- *; rewrite H3; auto.
exists (fun St => St = (MM1, (updatereg RR1 rB (RR1 rB + 1), S pc1))); split;
 auto.

constructor.
intros; compute in |- *; auto.
unfold curcmd, currf, curmem, curpc in |- *; intros (MM2, (RR2, pc2)); intros.
Exists2 1 TrivSP; split; compute in |- *; auto.

unfold curcmd, currf, curmem, curpc in |- *; intros (MM2, (RR2, pc2)); intros.
unfold Step in |- *; rewrite H5; auto.
exists (fun St => St = (MM2, (RR2, S pc2))); split; auto.
replace (blt_nat (RR2 rI) (RR2 rB)) with false; auto.
symmetry  in |- *; apply le_blt_false; auto.

apply wfcapcmd with rA rA rB 0; try tauto; try intros (MM3, (RR3, pc3)).
unfold curcmd, currf, curmem, curpc in |- *; intros.
unfold Step in |- *; rewrite H8; auto.
exists (fun St => St = (MM3, (updatereg RR3 rA (RR3 rA + RR3 rB), S pc3)));
 split; auto.

apply wfcapcmd with rA rA rA 0; try tauto; try intros (MM4, (RR4, pc4)).
unfold curcmd, currf, curmem, curpc in |- *; intros.
unfold Step in |- *; rewrite H10; auto.
exists (fun St => St = (MM4, (updatereg RR4 rA (MM4 (RR4 rA + 0)), S pc4)));
 split; auto.

constructor.
intros; compute in |- *; auto.
unfold curcmd, currf, curmem, curpc in |- *; intros (MM5, (RR5, pc5)); intros.
unfold Step in |- *; rewrite H12; auto.

generalize (H (MM,(RR,pc)) H0); intros (n,(Q,DX)); SplitAllHyps.
Exists2 n Q.
injection H11; intros _ e' e; rewrite e; rewrite e'; clear e e'; auto.
injection H9; intros _ e' e; rewrite e; rewrite e'; clear e e'; auto.
injection H7; intros _ e' e; rewrite e; rewrite e'; clear e e'; auto.
injection H4; intros _ e' e; rewrite e; rewrite e'; clear e e'; auto.
injection H2; intros _ e' e; rewrite e; rewrite e'; clear e e'; auto.
split; auto.

apply H14.
split.

unfold updatereg at 1.
replace (beq_regt r0 rA) with true; auto.
unfold updatereg at 1.
replace (beq_regt rA rA) with true; auto.
unfold updatereg at 1.
replace (beq_regt rA rB) with false; auto.
unfold updatereg at 1.
replace (beq_regt rA rI) with false; auto.
unfold updatereg at 1.
replace (beq_regt rB rB) with true; auto.
unfold updatereg at 1.
replace (beq_regt rB rI) with false; auto.
replace (RR rA + (RR rB + 1) + 0) with (S (RR r0) + RR r1); auto.
ring; unfold plus at 2.
replace (S (RR rB + RR rA)) with (S (RR rA) + RR rB); auto.
omega.

split.
unfold updatereg at 1; replace (beq_regt r2 rA) with false; auto.

split.
unfold updatereg at 1; replace (beq_regt r3 rA) with false; auto.

split.
unfold updatereg at 1; replace (beq_regt r4 rA) with false; auto.

split.
unfold updatereg at 1; replace (beq_regt r5 rA) with false; auto.

split. Qed.



Lemma Q_arrayget_cpinv :
  forall CLyt CT CH,
    let EC := ExtCode CLyt CH in
      IsCpGen EC CLyt CT CH ->
      let Q_arrayget:Pred := CpInv CLyt CH type_arrayget in
   (forall St, match St with (MM,(RR,pc)) =>
     (Q_arrayget St) -> 
     (exists n,exists Q, CT(RR r7)=someT(n,*Q) /\
       forall RR':RFile, 
	           (RR' r0)=(MM (S (RR r0)+(RR r1))) /\
	           (RR' r2)=(RR r2) /\
		   (RR' r3)=(RR r3) /\
		   (RR' r4)=(RR r4) /\
		   (RR' r5)=(RR r5) /\
		   (RR' r6)=(RR r6) ->
		   Q(MM,(RR',RR r7))) end).
Proof.
intros until Q_arrayget.
intros (MM,(RR,pc)); unfold Q_arrayget; clear Q_arrayget; 
  intros (DLyt,(Hp,(Rf,(HT,(Ts,(G,D)))))); SplitAllHyps.
inversion H0.
MCClear2 Ts0 F.
do 4
 (MCInvClear ListSubstL; ElimAllExistS; MCRewriteIn F ListSubstL; MCClear F;
   match goal with
   | H21:(_ = G0),H23:(_ = G) |- _ =>
       rewrite <- H21 (*..=G0*)  in H23 (*..=G*); MCClear G0;
        rename G1 into G0
   end).
MCInvClear ListSubstL; ElimAllExistS.
match goal with
| H23:(_ = G),H32:(_ = G0),H30:(F = _) |- _ =>
    rewrite H30 in H32; MCClear F; rewrite <- H32 in H23; MCClear G0
end.

assert (WFWordVal CH HT (Rf r7) (G r7)).
  apply H3 (*WFRegFile*); compute in |- *; auto.
MCInv WFWordVal; MCClear2 HT0 Ch.
MCRewriteIn G (intty = G r7); inversion H24 (*intty=...*).

(* JUMPING BACK TO TAL CODE *)
unfold IsCpGen in H; SplitAllHyps.
generalize (H4 (*TrCHeap*)  _ _ H24 (*fmaplook CH*)); intros;
 MCElimEx2 Ws a WordList.
generalize (H _ _ _ _ H24 H29); intros.
Exists2 (InstrSeq_size Is) (CpInv CLyt CH G0); split; auto.
replace (RR r7) with a; auto.
assert (TrWordVal CLyt DLyt (wl l) (RR rH)).
  unfold TrRegFile in H6; apply H6 with r7 (*TrRegFile*); auto.
MCInv TrWordVal; repeat MCUnfoldInSelf Lytlook; UnfoldFmaps;
 MCInjection (CLyt l).
MCInv ExtLytPred; SplitAllHyps.
generalize (H37 (*disjf*)  l); intros BAD; elim BAD;
 unfold indomf, fmaplook in |- *; split; eauto.

exists DLyt; exists Hp.

(* instantiate with the modified TAL register file, heap type, etc.
   and show that the modified real state translates to that ... *)
apply Super.

(* CASE WHERE JUMPING TO ANOTHER TYPE OF (i.e. CAP) CODE *)
apply Super.

(* nonsense case *)
MCRewriteIn G not; elim H24 (*intty=...*); compute in |- *; eauto.
Qed.


Lemma extcodewf_arraylib_alt
  : forall CLyt CH f,
    let EC := ExtCode_alt (CpInv CLyt CH type_arrayget) in
    forall Cs P,
    EC f = someT (Cs,*P) 
     -> forall CT, IsCpGen EC CLyt CT CH -> WFCapCmds TrivSP CT P Cs.
intros CLyt CH f EC; unfold EC, ExtCode_alt at 1.
induction f.

intros Cs P dinj; injection dinj; intros dinj1 dinj2.
MCRewrite2 Cs P; clear dinj dinj1 dinj2; clear Cs P.
intros.
constructor; intros; auto.
Exists2 1 TrivSP; split; auto.
MCUnfoldInSelf IsCpGen; SplitAllHyps.
assert (ExtCode_alt (CpInv CLyt CH type_arrayget) 0 = someT (_jd 0 :: nil,* TrivSP)).
  compute; auto.
match goal with 
 | H:ExtCode_alt _ _ = _ |- _ => generalize (H2 _ _ _ H) end;
compute; auto.

induction f.
intros Cs P dinj; injection dinj; intros dinj1 dinj2.
MCRewrite2 Cs P; clear dinj dinj1 dinj2; clear Cs P.
intros.
apply extcodewf_arrayget_alt; auto.

MCUnfoldInSelf IsCpGen; SplitAllHyps.
assert (ExtCode_alt (CpInv CLyt CH type_arrayget) 0 = someT (_jd 0 :: nil,* TrivSP)).
  compute; auto.
match goal with 
 | H:ExtCode_alt _ _ = _ |- _ => generalize (H0 _ _ _ H) end;
compute; auto.

(* here have to prove about CpInv *)
intros St.
apply Q_arrayget_cpinv; auto.

intros Cs P disc; discriminate disc.
Save.

Lemma XTALtoCAP_array_alt
  : forall CLyt CT DLyt CH H R Is St,
    let EC := ExtCode_alt (CpInv CLyt CH type_arrayget) in
    IsCpGen EC CLyt CT CH ->
    WFProgram (CH, (H, (R, Is))) ->
    TrProgram EC CLyt DLyt (CH, (H, (R, Is))) St ->
    WFCapState TrivSP St.
intros; apply (XTALtoCAP EC CLyt CT DLyt CH H R Is St); auto.
intros; apply extcodewf_arraylib_alt with CLyt CH f; auto.
Save.



Lemma XTALtoCAP_array 
  : forall CLyt CT DLyt CH H R Is St,
    let EC := ExtCode CLyt CH in
    IsCpGen EC CLyt CT CH ->
    WFProgram (CH, (H, (R, Is))) ->
    TrProgram EC CLyt DLyt (CH, (H, (R, Is))) St ->
    WFCapState TrivSP St.
intros; apply (XTALtoCAP EC CLyt CT DLyt CH H R Is St); auto.
intros; apply extcodewf_arraylib with CLyt CH f; auto.
Save.



Definition CLyt : LytType
  := fun l:label => match l with
                  | (1) => (Some 1) 
                  | _ => (None )
                    end.

Definition CH : CHeap
  := fun l:label => match l with
                  | (1) => Some (stub type_arrayget)
                  | _ => (None)
                    end.

Definition CT : CdSpec
  := fun l:Word =>
     match l with
     | (0) => someT (1 ,* TrivSP)
     | (1) => someT ((length cmds_arrayget) ,* (pred_arrayget CLyt CH))
     | _ => (noneT _)
     end.

Lemma goodCT : IsCpGen (ExtCode CLyt CH) CLyt CT CH.
split; auto.
unfold CT, ExtCode; intros.
induction w.
injection H; intros; MCRewrite2 Cs P.
induction w.
injection H; intros; MCRewrite2 Cs P.
AbsurdEq.

split; auto.
intros l; induction l.
unfold CH, CT, CLyt, Lytlook, fmaplook; intros; try AbsurdEq.
clear IHl; induction l; 
unfold CH, CT, CLyt, Lytlook, fmaplook; intros; try AbsurdEq.
discriminate H.

split; auto.
intros l; induction l.
unfold CH, CT, CLyt, Lytlook, fmaplook; intros; try AbsurdEq.
clear IHl; induction l; 
unfold CH, CT, CLyt, Lytlook, fmaplook; intros; try AbsurdEq.
injection H0; intros; MCRewrite w; unfold ExtCode; eauto.

split; auto.
intros w; induction w; try clear IHw.
unfold CT.
left.
injection H; intros; MCRewrite2 n P.
unfold ExtCode, CH, CT, CLyt, Lytlook, fmaplook; intros; try AbsurdEq.
exists (_jd 0 :: nil); split; compute; auto.

induction w; try clear IHw.
unfold CT.
left.
injection H; intros; MCRewrite2 n P.
unfold ExtCode, CH, CT, CLyt, Lytlook, fmaplook; intros; try AbsurdEq.
exists cmds_arrayget; split; unfold length, cmds_arrayget; split; auto.
unfold CpInterf; intros.
unfold pred_arrayget.
induction l; try AbsurdEq; try clear IHl.
induction l; try AbsurdEq; try clear IHl.
injection H4; intros; MCRewriteIn G CpInv.

unfold CT; intros; AbsurdEq.


induction l; try clear IHl; unfold Lytlook, fmaplook.
unfold CLyt; intros; AbsurdEq.

induction l; try clear IHL; unfold CLyt; intros; try AbsurdEq.
injection H; intros; MCRewrite w; compute; eauto.
Save.


Lemma XTALtoCAP_array_example
  : forall DLyt H R Is St,
    let EC := ExtCode CLyt CH in
    WFProgram (CH, (H, (R, Is))) ->
    TrProgram EC CLyt DLyt (CH, (H, (R, Is))) St ->
    WFCapState TrivSP St.
intros; apply XTALtoCAP_array with CLyt CT DLyt CH H R Is; auto.
apply goodCT.
Save.


(**************************************************************************)
(**************************************************************************)
(**************************************************************************)
(**************************************************************************)








(**************** REMARK ::: *********************)
(* Why it seems we can't use '=' (i.e. Coq's "eq" equality) to
   handle equivalence of code types throughout substitution, lifting, etc. *)

Definition G : RegFileTy := fun r => intty.
Definition T : CodeTy := tabsct (basectl 1 (fun r => varL _)).

Goal InstCodeTy T (intty :: nil) G.
assert
 (ListSubstL 0 (basectl 1 (fun _ : Reg => varL _)) (
    intty :: nil) (fun _ : Reg => inttyL)).
  generalize (lsub_base 0 (fun _ : Reg => varL _) intty).
  unfold SubstL, subst_aux in |- *.
  EQREC (eq_rec 0 OmegaL inttyL 0 (myeqaddS 0 0 (refl_equal 1))) inttyL.
generalize
 (instct_tabs (basectl 1 (fun r => varL _)) (intty :: nil) (
    fun r => inttyL) H).
compute in |- *; intros; auto.
Qed.

(** Now a more complicated ... *)

Definition G' : RegFileTy :=
  fun r => match r with
           | r0 => arrty
           | _ => intty
           end.
Definition T' : CodeTy :=
  tabsct
    (basectl 1
       (fun r => match r with
                 | r0 => varL _
                 | _ => liftL _ inttyL
                 end)).

Goal InstCodeTy T (arrty :: nil) G.
assert
 (ListSubstL 0
    (basectl 1
       (fun r : Reg => match r with
                       | r0 => varL _
                       | _ => liftL _ inttyL
                       end)) (arrty :: nil)
    (fun r : Reg => match r with
                    | r0 => arrtyL
                    | _ => inttyL
                    end)).
  generalize
   (lsub_base 0
      (fun r => match r with
                | r0 => varL _
                | _ => liftL _ inttyL
                end) arrty).
    unfold SubstL, subst_aux in |- *; fold subst_aux in |- *.
    intros.
 (* notice at this point, we need to sort of do a case analysis on the 
    [r:Reg] in H so that we can reduce subst_aux to get the desired 
    conclusion. But I don't see it can be done ... *)
apply Super.
apply Super.
Save.
