(* 
  Coq code for the paper:
  Interfacing Hoare Logic and Type Systems for Foundational Proof-Carrying Code
  by Nadeem Abdul Hamid and Zhong Shao

  Coq version 8.0

  This code is a prototype version only -- proofs are mostly complete.
*)

(*
 *
 Array-Extensible-TAL-with-polymorphism

 A version of TAL supporting primitive array types
 but no operations on them!

 A revised version of (atal.v) separating code and data heaps

 For Coq Version 8.0
*)


Require Import Eqdep.
Definition eq_refl := UIP_refl.
Ltac EQREC a b := replace a with b; try apply eq_rec_eq; auto.
Ltac ElimExistS h := generalize (inj_pair2 _ _ _ _ _ h); clear h; intros.
Ltac ElimAllExistS :=
  repeat
   match goal with
   | h:(existS _ _ _ = existS _ _ _) |- _ => ElimExistS h
   end.

Require Import Arith.
Require Import ArithRing.
Require Import EqNat.
Require Import Omega.
Require Import mystuff.
Require Import tis.
Require Import captis.
Require Import fmap.

(**************************************************************************)
(**************************************************************************)
(**************************************************************************)
(*** aXTALp SYNTAX ***)

Notation label := nat.
Notation int := label.

Require Import talreg.  (* r0 ... r7  *)


(* lifted types *)
Inductive OmegaL : int -> Set :=
  | varL : forall i : int, OmegaL (S i)
  | liftL : forall i : int, OmegaL i -> OmegaL (S i)
  | inttyL : OmegaL 0
  | arrtyL : OmegaL 0
  | codetyL : forall i : int, CodeTyL i -> OmegaL i
with CodeTyL : int -> Set :=
  | basectl : forall i : int, (Reg -> OmegaL i) -> CodeTyL i
  | tabsctl : forall i : int, CodeTyL (S i) -> CodeTyL i.



(* The actual types - no deBruijn stuff at the top level *)
Inductive Omega : Set :=
  | intty : Omega
  | arrty : Omega
  | codety : CodeTy -> Omega
with CodeTy : Set :=
  | basect : (Reg -> Omega) -> CodeTy
  | tabsct : CodeTyL 1 -> CodeTy.


Inductive oleq : forall i : int, OmegaL i -> OmegaL i -> Prop :=
  | oleq_refl : forall i F, oleq i F F
  | oleq_lift :
      forall i (F F' : OmegaL i),
      oleq i F F' -> oleq (S i) (liftL _ F) (liftL _ F')
  | oleq_code :
      forall i F F', octleq _ F F' -> oleq i (codetyL _ F) (codetyL _ F')
with octleq : forall i : int, CodeTyL i -> CodeTyL i -> Prop := (*octleq_refl : (i,F:?)(octleq i F F)*)
  | octleq_base :
      forall i F F',
      (forall r, oleq i (F r) (F' r)) ->
      octleq i (basectl _ F) (basectl _ F')
  | octleq_tabs :
      forall i F F',
      octleq (S i) F F' -> octleq i (tabsctl _ F) (tabsctl _ F')
with oeq : Omega -> Omega -> Prop :=
  | oeq_refl : forall T, oeq T T
  | oeq_code : forall F F', octeq F F' -> oeq (codety F) (codety F')
with octeq : CodeTy -> CodeTy -> Prop := (*octeq_refl : (F:?)(octeq F F)*)
  | octeq_base :
      forall F F',
      (forall r, oeq (F r) (F' r)) -> octeq (basect F) (basect F')
  | octeq_tabs : forall F F', octleq _ F F' -> octeq (tabsct F) (tabsct F').

Scheme oleq_octleq_ind := Minimality for oleq
  Sort Prop
  with octleq_oleq_ind := Minimality for octleq Sort Prop.

Scheme oeq_octeq_ind := Minimality for oeq Sort Prop
  with octeq_oeq_ind := Minimality for octeq Sort Prop.

Lemma sym_oleq : forall i T T', oleq i T T' -> oleq i T' T.
intros i T T' H.
generalize H.
elim H using
 oleq_octleq_ind
  with
    (fun i T T' => oleq i T T' -> oleq i T' T)
    (fun i F F' => octleq i F F' -> octleq i F' F)
    i
    T
    T'; intros; auto; try constructor; auto.
Qed.

Lemma sym_octleq : forall i T T', octleq i T T' -> octleq i T' T.
intros i T T' H.
generalize H.
elim H using
 octleq_oleq_ind
  with
    (fun i T T' => oleq i T T' -> oleq i T' T)
    (fun i F F' => octleq i F F' -> octleq i F' F)
    i
    T
    T'; intros; auto; try constructor; auto.
Qed.

Lemma sym_octeq : forall F F', octeq F F' -> octeq F' F.
intros F F' H.
generalize H;
 elim H using
  octeq_oeq_ind
   with
     (fun T T' => oeq T T' -> oeq T' T)
     (fun F F' => octeq F F' -> octeq F' F)
     F
     F'; intros; auto; try constructor; auto.
apply sym_octleq; auto.
Qed.

Lemma sym_oeq : forall F F', oeq F F' -> oeq F' F.
intros F F' H.
generalize H;
 elim H using
  oeq_octeq_ind
   with
     (fun T T' => oeq T T' -> oeq T' T)
     (fun F F' => octeq F F' -> octeq F' F)
     F
     F'; intros; auto; try constructor; auto.
apply sym_octleq; auto.
Qed.

Lemma octleq_refl : forall i F, octleq i F F.
induction F; constructor; intros; auto.
apply oleq_refl.
Qed.

Lemma octeq_refl : forall F, octeq F F.
induction F; constructor; intros; auto.
apply oeq_refl.
apply octleq_refl.
Qed.

Hint Immediate sym_oleq sym_octleq sym_octeq sym_oeq oeq_refl oleq_refl
  octeq_refl octleq_refl.

Lemma trans_octleq :
 forall i F F',
 octleq i F F' -> forall F'', octleq _ F' F'' -> octleq _ F F''.
intros i F F' H.
elim H using
 octleq_oleq_ind
  with
    (fun i T T' => forall T'', oleq i T' T'' -> oleq _ T T'')
    (fun i F F' => forall F'', octleq i F' F'' -> octleq _ F F'')
    i
    F
    F'; intros; auto.
MCInv oleq. 
  replace T'' with (liftL _ F'0);
   [ constructor; auto
   | generalize (inj_pair2 _ _ _ _ _ H6); intros; MCRewrite T'';
      symmetry  in |- *; apply (inj_pair2 _ _ _ _ _ H5); 
      auto ].
  replace T'' with (liftL _ F'1); [ idtac | apply (inj_pair2 _ _ _ _ _ H5) ].
  generalize (inj_pair2 _ _ _ _ _ H4); intros;
   MCRewriteIn F1 (oleq i0 F1 F'1); constructor; auto.

MCInv oleq.
  generalize (inj_pair2 _ _ _ _ _ H6); intros;
   generalize (inj_pair2 _ _ _ _ _ H5); intros; MCRewrite T''; 
   MCRewrite F1; constructor; auto.
  generalize (inj_pair2 _ _ _ _ _ H3); intros;
   generalize (inj_pair2 _ _ _ _ _ H5); intros; MCRewrite T'';
   MCRewriteIn F1 (octleq i0 F1 F'1); constructor; 
   auto.

MCInv octleq.
  generalize (inj_pair2 _ _ _ _ _ H3); intros;
   generalize (inj_pair2 _ _ _ _ _ H5); intros; MCRewrite F''; 
   constructor; intros; apply H1; MCRewrite F'0; auto.

MCInv octleq.
  generalize (inj_pair2 _ _ _ _ _ H3); intros;
   generalize (inj_pair2 _ _ _ _ _ H5); intros; MCRewrite F''; 
   constructor; intros; apply H1; MCRewrite F'0; auto.
Qed.

Lemma trans_oeq :
 forall T T', oeq T T' -> forall T'', oeq T' T'' -> oeq T T''.
intros T T' H.
elim H using
 oeq_octeq_ind
  with
    (fun T T' => forall T'', oeq T' T'' -> oeq T T'')
    (fun F F' => forall F'', octeq F' F'' -> octeq F F'')
    T
    T'; intros; auto.
MCInv oeq; try constructor; auto.
MCInv octeq; try constructor; auto.
MCInv octeq; constructor; auto.
apply trans_octleq with F'; auto.
Qed.


Lemma trans_octeq :
 forall T T', octeq T T' -> forall T'', octeq T' T'' -> octeq T T''.
intros T T' H.
elim H using
 octeq_oeq_ind
  with
    (fun T T' => forall T'', oeq T' T'' -> oeq T T'')
    (fun F F' => forall F'', octeq F' F'' -> octeq F F'')
    T
    T'; intros; auto.
MCInv oeq; try constructor; auto.
MCInv octeq; try constructor; auto.
MCInv octeq; constructor; auto.
apply trans_octleq with F'; auto.
Qed.




(**************************************************************************)
(* Utilities for deBruijn indices *)
(* substitution *)


Let myeqaddS : forall n m : int, S n = S m -> n = m.
intros; injection H.
intros; assumption.
Defined.

Fixpoint subst_aux (i : int) (t : OmegaL i) {struct t} :
 forall j : int, i = S j -> OmegaL j -> OmegaL j :=
  match
    t as X in (OmegaL i)
    return (forall (j : int) (p : i = S j) (e' : OmegaL j), OmegaL j)
  with
  | varL n => fun (j : int) (_ : S n = S j) (e' : OmegaL j) => e'
  | liftL n t0 =>
      fun (j : int) (p : S n = S j) (_ : OmegaL j) =>
      eq_rec n OmegaL t0 j (myeqaddS n j p)
  | inttyL =>
      fun (j : int) (_ : 0 = S j) (e' : OmegaL j) => e' (* not reached *)
  | arrtyL =>
      fun (j : int) (_ : 0 = S j) (e' : OmegaL j) => e' (* not reached *)
  | codetyL n G =>
      fun (j : int) (p : n = S j) (e' : OmegaL j) =>
      codetyL j (subst_codety_aux n G j p e')
  end
 
 with subst_codety_aux (i : int) (t : CodeTyL i) {struct t} :
 forall j : int, i = S j -> OmegaL j -> CodeTyL j :=
  match
    t in (CodeTyL i)
    return (forall j : int, i = S j -> OmegaL j -> CodeTyL j)
  with
  | basectl n G =>
      fun (j : int) (p : n = S j) (e' : OmegaL j) =>
      basectl j (fun r => subst_aux n (G r) j p e')
  | tabsctl n G =>
      fun (j : int) (p : n = S j) (e' : OmegaL j) =>
      tabsctl j (subst_codety_aux (S n) G (S j) (f_equal S p) (liftL j e'))
  end.

Definition SubstL (n : int) (T : OmegaL (S n)) :=
  subst_aux (S n) T n (refl_equal (S n)).

Definition SubstCodeTyL (n : int) (T : CodeTyL (S n)) :=
  subst_codety_aux (S n) T n (refl_equal (S n)).

(* conversion from the real Omega *)
Fixpoint RLift (T : Omega) : OmegaL 0 :=
  match T with
  | intty => inttyL
  | arrty => arrtyL
  | codety G => codetyL _ (RLiftCodeTy G)
  end
 
 with RLiftCodeTy (G : CodeTy) : CodeTyL 0 :=
  match G with
  | basect G => basectl _ (fun r => RLift (G r))
  | tabsct G => tabsctl _ G
  end.

(* conversion to the real Omega *)
Fixpoint RUnlift i (T : OmegaL i) {struct T} : i = 0 -> Omega :=
  match T in (OmegaL n) return (n = 0 -> Omega) with
  | varL i =>
      fun p : S i = 0 =>
      False_rec Omega
        (eq_ind (S i)
           (fun e : int => match e with
                           | O => False
                           | _ => True
                           end) I 0 p)
  | liftL i T0 =>
      fun p : S i = 0 =>
      False_rec Omega
        (eq_ind (S i)
           (fun e : int => match e with
                           | O => False
                           | _ => True
                           end) I 0 p)
  | inttyL => fun _ : 0 = 0 => intty
  | arrtyL => fun _ : 0 = 0 => arrty
  | codetyL i G =>
      fun p : i = 0 =>
      codety (RUnliftCodeTy _ (eq_rec i _ G 0 p) (refl_equal _))
  end
 
 with RUnliftCodeTy i (G : CodeTyL i) {struct G} : 
 i = 0 -> CodeTy :=
  match G in (CodeTyL n) return (n = 0 -> CodeTy) with
  | basectl i G =>
      fun p : i = 0 =>
      basect (fun r => RUnlift _ (eq_rec i _ (G r) 0 p) (refl_equal _))
  | tabsctl i G => fun p : i = 0 => tabsct (eq_rec (S i) _ G 1 (eq_S i 0 p))
  end.

Definition OO : 0 = 0 := refl_equal _.

Definition liftLto : forall k i j : int, j = k + i -> OmegaL i -> OmegaL j.
induction k; intros.
MCRewrite j; auto.
replace j with (S (k + i)).
apply liftL; apply IHk with i; auto.
Defined.

(* built-in lemma is not opaque, we need this one to compute (reduce) *)
Definition plusn0 : forall n, n=n+0.
induction n; intros.
unfold plus; apply refl_equal.
unfold plus.
apply (f_equal  S).
exact IHn.
Defined.

Definition liftLzeroto : forall i : int, OmegaL 0 -> OmegaL i.
intros; apply liftLto with i 0.
apply plusn0.
exact H.
Defined.


(**************************** ENOUGH OF THAT! ******************************)
(**************************************************************************)

Definition Omegalist := list Omega.
Definition RegFileTy := Reg -> Omega.
Definition RegFileTyL i := Reg -> OmegaL i.

Inductive WordVal : Set :=
  | wl : int -> WordVal
  | wi : int -> WordVal.

Inductive Instr : Set :=
  | add : Reg -> Reg -> Reg -> Instr
  | movi : Reg -> int -> Instr
  | movl : Reg -> int -> Instr.

Inductive InstrSeq : Set :=
  | iseq : Instr -> InstrSeq -> InstrSeq
  | jd : int -> Omegalist -> InstrSeq
  | jmp : Reg -> Omegalist -> InstrSeq.

Inductive CHeapVal : Set :=
  | code : CodeTy -> InstrSeq -> CHeapVal
  | stub : CodeTy -> CHeapVal.

Definition typeofchv (chv : CHeapVal) :=
  match chv with
  | code G _ => G
  | stub G => G
  end.

Inductive HeapVal : Set :=
    array : list int -> HeapVal.

Definition RegFile := Reg -> WordVal.
Definition CHeap := fmap int CHeapVal.
(* no "cheap type" -- use the same CHeap, it doesn't ever change *)

Definition Heap := fmap int HeapVal.
Definition HeapTy := fmap int Omega.

Definition SubProgram := (Heap * (RegFile * InstrSeq))%type.
Definition Program := (CHeap * SubProgram)%type.


Definition cheapof (P : Program) := match P with
                                    | (Ch, _) => Ch
                                    end.
Definition dheapof (P : Program) := match P with
                                    | (_, (DH, _)) => DH
                                    end.
Definition regfileof (P : Program) :=
  match P with
  | (_, (_, (R, _))) => R
  end.
Definition iseqof (P : Program) := match P with
                                   | (_, (_, (_, Is))) => Is
                                   end.


(**************************************************************************)
(**************************************************************************)
(*** iXTAL SEMANTICS ***)

(* tail subsets of instruction sequences *)
(*   e.g.  I <= (iseq i (iseq i' I)) @ 2   *)
Inductive SubseqIS : InstrSeq -> InstrSeq -> int -> Prop :=
  | isubdO : forall I, SubseqIS I I 0
  | isubdS : forall I I' i n, SubseqIS I' I n -> SubseqIS I' (iseq i I) (S n).

Definition rfupd (R : RegFile) (r : Reg) (v : WordVal) : RegFile :=
  fun r' => if beq_reg r r' then v else R r'.
Definition rftupd (G : RegFileTy) r t : RegFileTy :=
  fun r' => if beq_reg r r' then t else G r'.

Definition hupd (H : Heap) l hv : Heap :=
  fun l' => if beq_nat l l' then Some hv else H l'.
Definition htupd (HT : HeapTy) l t : HeapTy :=
  fun l' => if beq_nat l l' then Some t else HT l'.

Section Flook.
Set Implicit Arguments.
Unset Strict Implicit.
Definition flook (A : Set) (M : A -> Omega) a b := M a = b.
Set Strict Implicit.
Unset Implicit Arguments.
End Flook.




(***************************************************************************)
(* Operational Semantics *)
Inductive Eval : CHeap -> SubProgram -> SubProgram -> Prop :=
  | ev_add :
      forall CH H R Is rd rs rt s t,
      let i := add rd rs rt in
      let R' := rfupd R rd (wi (s + t)) in
      R rs = wi s -> R rt = wi t -> Eval CH (H, (R, iseq i Is)) (H, (R', Is))
  | ev_movi :
      forall CH H R Is rd n,
      let i := movi rd n in
      let R' := rfupd R rd (wi n) in
      Eval CH (H, (R, iseq i Is)) (H, (R', Is))
  | ev_movl :
      forall CH H R Is rd l,
      let i := movl rd l in
      let R' := rfupd R rd (wl l) in
      Eval CH (H, (R, iseq i Is)) (H, (R', Is))
  | ev_jd :
      forall CH H R l Ts G Is,
      fmaplook CH l (code G Is) -> Eval CH (H, (R, jd l Ts)) (H, (R, Is))
  | ev_jmp :
      forall CH H R r l Ts G Is,
      R r = wl l ->
      fmaplook CH l (code G Is) -> Eval CH (H, (R, jmp r Ts)) (H, (R, Is)).

(***************************************************************************)
(* Static Semantics *)

(* word values *)
Inductive WFWordVal : CHeap -> HeapTy -> WordVal -> Omega -> Prop :=
  | int_wval : forall Ch HT i, WFWordVal Ch HT (wi i) intty
  | label_wval0 :
      forall Ch HT l G G' Is,
      fmaplook Ch l (code G Is) ->
      octeq G G' -> WFWordVal Ch HT (wl l) (codety G')
  | label_wval1 :
      forall Ch HT l G G',
      fmaplook Ch l (stub G) ->
      octeq G G' -> WFWordVal Ch HT (wl l) (codety G')
  | label_wval2 :
      forall Ch HT l t,
      fmaplook HT l t ->
      ~ (exists T : _, t = codety T) -> WFWordVal Ch HT (wl l) t.



(* instruction sequences *)
Inductive ListSubstL :
forall n, CodeTyL (S n) -> Omegalist -> (Reg -> OmegaL n) -> Prop :=
  | lsub_base :
      forall i F t,
      ListSubstL _ (basectl (S i) F) (t :: nil)
        (fun r => SubstL _ (F r) (liftLzeroto _ (RLift t)))
  | lsub_tabs :
      forall i F t Ts G,
      ListSubstL (S i) F Ts G ->
      ListSubstL _ (tabsctl (S i) F) (t :: Ts)
        (fun r => SubstL _ (G r) (liftLzeroto _ (RLift t))).

Inductive InstCodeTy : CodeTy -> Omegalist -> RegFileTy -> Prop :=
  | instct_base : forall F, InstCodeTy (basect F) nil F
  | instct_tabs :
      forall F Ts G,
      ListSubstL 0 F Ts G ->
      InstCodeTy (tabsct F) Ts (fun r => RUnlift _ (G r) OO).

Inductive WFInstrSeq : CHeap -> RegFileTy -> InstrSeq -> Prop :=
  | add_i :
      forall Ch G rd rs rt Is,
      flook G rs intty ->
      flook G rt intty ->
      WFInstrSeq Ch (rftupd G rd intty) Is ->
      WFInstrSeq Ch G (iseq (add rd rs rt) Is)
  | movi_i :
      forall Ch G rd i Is,
      WFInstrSeq Ch (rftupd G rd intty) Is ->
      WFInstrSeq Ch G (iseq (movi rd i) Is)
  | movl_i :
      forall Ch G rd l Is hv T,
      fmaplook Ch l hv ->
      typeofchv hv = T ->
      WFInstrSeq Ch (rftupd G rd (codety T)) Is ->
      WFInstrSeq Ch G (iseq (movl rd l) Is)
  | jd_i :
      forall Ch G G' l Ts hv T,
      fmaplook Ch l hv ->
      typeofchv hv = T ->
      InstCodeTy T Ts G' ->
      octeq (basect G) (basect G') -> WFInstrSeq Ch G (jd l Ts)
  | jmp_i :
      forall Ch G G' r Ts T,
      flook G r (codety T) ->
      InstCodeTy T Ts G' ->
      octeq (basect G) (basect G') -> WFInstrSeq Ch G (jmp r Ts).

(* heap values *)
Inductive WFCHeapVal : CHeap -> CHeapVal -> Prop :=
  | code_hval :
      forall Ch T Is,
      (forall Ts G, InstCodeTy T Ts G -> WFInstrSeq Ch G Is) ->
      WFCHeapVal Ch (code T Is)
  | stub_hval : forall Ch T, WFCHeapVal Ch (stub T).

Inductive WFHeapVal : HeapTy -> HeapVal -> Omega -> Prop :=
    array_hval : forall HT Ns, WFHeapVal HT (array Ns) arrty.

(* top-level rules *)
Definition WFRegFile (Ch : CHeap) (HT : HeapTy) (R : RegFile) 
  (G : RegFileTy) : Prop :=
  forall r t, flook G r t -> WFWordVal Ch HT (R r) t.

Definition WFCHeap (Ch : CHeap) : Prop :=
  (exists Lim : _, (forall l, Lim <= l -> notindomf Ch l)) /\
  (forall l chv, fmaplook Ch l chv -> WFCHeapVal Ch chv).

Definition WFHeap (H : Heap) (HT : HeapTy) : Prop :=
  (exists Lim : _, (forall l, Lim <= l -> notindomf HT l)) /\
  eqdomf _ _ _ H HT /\
  (forall l h t, fmaplook H l h -> fmaplook HT l t -> WFHeapVal HT h t).


Definition InstrSeqInHeap (Is : InstrSeq) (Ch : CHeap) : Prop :=
  exists l : _,
    (exists T : _,
       (exists Is' : _,
          (exists n : _, fmaplook Ch l (code T Is') /\ SubseqIS Is Is' n))).

Inductive WFProgram : Program -> Prop :=
    prog_top :
      forall G Ch H HT R Is,
      WFCHeap Ch ->
      WFHeap H HT ->
      WFRegFile Ch HT R G ->
      WFInstrSeq Ch G Is ->
      InstrSeqInHeap Is Ch -> WFProgram (Ch, (H, (R, Is))).

(*************************************************************************)
(*************************************************************************)
(*************************************************************************)
(*************************************************************************)
(* iXTAL PRESERVATION / PROGRESS THEOREMS *)


Lemma SubseqIS_iseq :
 forall Is' i Is n, SubseqIS (iseq i Is) Is' n -> SubseqIS Is Is' (S n).
induction Is'; intros; MCInv SubseqIS;
 [ constructor; constructor | constructor; apply IHIs' with i0; auto ].
Qed.


Lemma InstrSeqInHeap_tail :
 forall i Is H, InstrSeqInHeap (iseq i Is) H -> InstrSeqInHeap Is H.
unfold InstrSeqInHeap in |- *; intros i Is H.
intros (l, (G, (Is', (n, HH)))); SplitAllHyps.
Exists4 l G Is' (S n); split; auto.
apply SubseqIS_iseq with i; auto.
Qed.

Definition RegHeapTy (HT : HeapTy) := forall l, HT l <> Some intty.

Lemma wfheap_regheapty : forall H HT, WFHeap H HT -> RegHeapTy HT.
unfold WFHeap, fmaplook in |- *; IntrosSplitH.
intro; intro.
generalize (H0 l); unfold iff, indomf, fmaplook in |- *; IntrosSplitH.
elim H6; eauto; intros.
generalize (H3 _ _ _ H4 H2); intros HWF.
inversion HWF.
Qed.


(* NOTE: we never actually need to use progress_jd and
         preserv_jd for the compilation to CAP ... *)
Lemma progress_jd :
 forall G HT CH H R l Ts (P:=(H, (R, jd l Ts))) (D1 : WFCHeap CH)
   (D2 : WFHeap H HT) (D3 : WFRegFile CH HT R G)
   (D4 : WFInstrSeq CH G (jd l Ts))
   (D5 : ~ (exists T : _, fmaplook CH l (stub T))),
 exists P', Eval CH P P'.
intros.
cut (P=P).        (* experimenting with coq 8.0 tactics :) *)
 2:auto.
 1:intros DP.
unfold P in  D1, DP |- *; try clear P.
MCInv WFInstrSeq.
generalize D1; intros (D1a, D1b).
generalize (D1b _ _ H2); intros.
MCInv WFCHeapVal.
exists (H, (R, Is)).
apply ev_jd with T0; auto.
MCRewriteRev hv.
elim D5; exists T0; MCRewriteRev hv.
Qed.




Lemma octleq_octeq :
 forall i (D : i = 0) F F',
 octleq i (basectl _ F) (basectl _ F') ->
 octeq (basect (fun r => RUnlift _ (F r) D))
   (basect (fun r => RUnlift _ (F' r) D)).
intros.
assert
 (octeq (RUnliftCodeTy _ (basectl _ F) D) (RUnliftCodeTy _ (basectl _ F') D)).
generalize D.
elim H using
 octleq_oleq_ind
  with
    (fun i T T' => forall D : i = 0, oeq (RUnlift i T D) (RUnlift _ T' D))
    (fun i F F' =>
     forall D : i = 0, octeq (RUnliftCodeTy i F D) (RUnliftCodeTy _ F' D))
    i
    (basectl i F)
    (basectl i F'); intros; auto.
generalize (H1 D0); intros HH.
inversion HH; UnfoldFold RUnlift; fold RUnliftCodeTy in |- *.


generalize H4 H3 H2.
generalize D0.
generalize F0 F'0.
MCRewrite i0.
intros.
generalize (eq_refl _ _ D1); intros; MCRewriteRev D1.
EQREC (eq_rec 0 CodeTyL F2 0 D1) F2.
EQREC (eq_rec 0 CodeTyL F'2 0 D1) F'2.
rewrite <- H6; rewrite <- H7.
constructor.
constructor; auto.

generalize D0 F0 F'0 HH.
MCRewrite i0; intros.
generalize (eq_refl _ _ D1); intros; MCRewriteRev D1.
EQREC (eq_rec 0 CodeTyL F2 0 D1) F2.
EQREC (eq_rec 0 CodeTyL F'2 0 D1) F'2.
constructor; auto.

UnfoldFold RUnliftCodeTy; fold RUnlift in |- *; constructor; intros.
generalize D0 F0 F'0 H1.
MCRewrite i0; intros.
generalize (eq_refl _ _ D1); intros; MCRewriteRev D1.
EQREC (eq_rec 0 OmegaL (F1 r) 0 D1) (F1 r).
EQREC (eq_rec 0 OmegaL (F'1 r) 0 D1) (F'1 r).

UnfoldFold RUnliftCodeTy; fold RUnlift in |- *; constructor; intros.
generalize D0.
generalize F0 F'0 H0.
rewrite D0; intros.
EQREC (eq_rec 1 CodeTyL F1 1 (eq_S 0 0 D1)) F1.
EQREC (eq_rec 1 CodeTyL F'1 1 (eq_S 0 0 D1)) F'1.

generalize H0; unfold RUnliftCodeTy in |- *; fold RUnlift in |- *.
generalize F F' D.
rewrite D; intros.
inversion H1.
constructor; intros.
generalize (eq_refl _ _ D0); intros ee; rewrite <- ee in H4.
generalize (H4 r).
EQREC (eq_rec 0 OmegaL (F0 r) 0 D0) (F0 r).
EQREC (eq_rec 0 OmegaL (F'0 r) 0 D0) (F'0 r).
Qed.


Definition omegal_conv : forall i j, i = j -> OmegaL i -> OmegaL j.
intros; MCRewrite j.
Defined.

Definition codetyl_conv : forall i j, i = j -> CodeTyL i -> CodeTyL j.
intros; MCRewrite j.
Defined.

Lemma oleq_subst :
 forall j t t',
 oleq j t t' ->
 forall i (D : j = S i) a,
 oleq i (SubstL i (omegal_conv _ _ D t) a)
   (SubstL i (omegal_conv _ _ D t') a).
intros j t t' H.

elim H using
 oleq_octleq_ind
  with
    (fun j (T T' : OmegaL j) =>
     forall i (D : j = S i) (a : OmegaL i),
     oleq i (SubstL i (omegal_conv _ _ D T) a)
       (SubstL i (omegal_conv _ _ D T') a))
    (fun j (F F' : CodeTyL j) =>
     forall i (D : j = S i) (a : OmegaL i),
     octleq i (SubstCodeTyL i (codetyl_conv _ _ D F) a)
       (SubstCodeTyL i (codetyl_conv _ _ D F') a))
    j
    t
    t'; intros; auto.
unfold omegal_conv in |- *.
injection D; intros.
generalize H0.
generalize F F' a.
generalize D.
MCRewrite i0; intros.
EQREC (eq_rec (S i) (fun n : int => OmegaL n) (liftL i F0) (S i) D0)
 (liftL i F0); try apply eq_rec_eq with (P := fun n : int => OmegaL n).
EQREC (eq_rec (S i) (fun n : int => OmegaL n) (liftL i F'0) (S i) D0)
 (liftL i F'0); try apply eq_rec_eq with (P := fun n : int => OmegaL n).

unfold omegal_conv in |- *.
generalize (H1 _ D a).
generalize F F' a.
generalize D. 
MCRewrite i.
intros D0 F0 F'0 a0.
unfold codetyl_conv in |- *.
EQREC (eq_rec (S i0) (fun n : int => CodeTyL n) F0 (S i0) D0) F0;
 try apply eq_rec_eq with (P := fun n : int => CodeTyL n).
EQREC (eq_rec (S i0) (fun n : int => CodeTyL n) F'0 (S i0) D0) F'0;
 try apply eq_rec_eq with (P := fun n : int => CodeTyL n).
unfold SubstCodeTyL in |- *.
intros.
EQREC (eq_rec (S i0) (fun n : int => OmegaL n) (codetyL (S i0) F0) (S i0) D0)
 (codetyL (S i0) F0); try apply eq_rec_eq with (P := fun n : int => OmegaL n).
EQREC
 (eq_rec (S i0) (fun n : int => OmegaL n) (codetyL (S i0) F'0) (S i0) D0)
 (codetyL (S i0) F'0);
 try apply eq_rec_eq with (P := fun n : int => OmegaL n).
unfold SubstL, subst_aux in |- *; fold subst_codety_aux in |- *.
constructor; auto.

unfold codetyl_conv in |- *.
generalize H1.
generalize F F' a.
generalize D. 
clear a.
MCRewrite i; intros.
EQREC
 (eq_rec (S i0) (fun n : int => CodeTyL n) (basectl (S i0) F0) (S i0) D0)
 (basectl (S i0) F0);
 try apply eq_rec_eq with (P := fun n : int => CodeTyL n).
EQREC
 (eq_rec (S i0) (fun n : int => CodeTyL n) (basectl (S i0) F'0) (S i0) D0)
 (basectl (S i0) F'0);
 try apply eq_rec_eq with (P := fun n : int => CodeTyL n).
unfold SubstCodeTyL, subst_codety_aux in |- *; fold subst_aux in |- *;
 constructor.
intros.
generalize (H2 r _ D0 a); unfold omegal_conv in |- *.
EQREC (eq_rec (S i0) (fun n : int => OmegaL n) (F0 r) (S i0) D0) (F0 r);
 try apply eq_rec_eq with (P := fun n : int => OmegaL n).
EQREC (eq_rec (S i0) (fun n : int => OmegaL n) (F'0 r) (S i0) D0) (F'0 r);
 try apply eq_rec_eq with (P := fun n : int => OmegaL n).

unfold codetyl_conv in |- *.
generalize H1; generalize F F' a; generalize D.
clear a; rewrite D; clear D; intros.
EQREC (eq_rec (S i0) (fun n : int => CodeTyL n) (tabsctl (S i0) F0) (S i0) D)
 (tabsctl (S i0) F0);
 try apply eq_rec_eq with (P := fun n : int => CodeTyL n).
EQREC
 (eq_rec (S i0) (fun n : int => CodeTyL n) (tabsctl (S i0) F'0) (S i0) D)
 (tabsctl (S i0) F'0);
 try apply eq_rec_eq with (P := fun n : int => CodeTyL n).
unfold SubstCodeTyL, subst_codety_aux in |- *; fold subst_codety_aux in |- *.
constructor.
generalize (H2 (S i0) (eq_S _ _ D) (liftL _ a)).
unfold codetyl_conv in |- *.
EQREC
 (eq_rec (S (S i0)) (fun n : int => CodeTyL n) F0 (
    S (S i0)) (eq_S (S i0) (S i0) D)) F0;
 try apply eq_rec_eq with (P := fun n : int => CodeTyL n).
EQREC
 (eq_rec (S (S i0)) (fun n : int => CodeTyL n) F'0 (
    S (S i0)) (eq_S (S i0) (S i0) D)) F'0;
 try apply eq_rec_eq with (P := fun n : int => CodeTyL n).
Qed.


Lemma octleq_listsubstl :
 forall Ts i F F' G,
 ListSubstL i F Ts G ->
 octleq _ F F' ->
 exists G' : _,
   ListSubstL i F' Ts G' /\ octleq _ (basectl _ G) (basectl _ G').
induction Ts; intros.
inversion H.
inversion H.
ElimAllExistS.
MCRewriteIn F octleq.
inversion H0.
ElimAllExistS.
MCRewrite F'.
exists (fun r => SubstL _ (F'0 r) (liftLzeroto i (RLift a))).
split; auto.
constructor.
MCRewrite G.
constructor; intros.
generalize (H9 r); intros.
change
  (oleq i
     (SubstL i (omegal_conv (S i) (S i) (refl_equal _) (F0 r))
        (liftLzeroto i (RLift a)))
     (SubstL i (omegal_conv (S i) (S i) (refl_equal _) (F'0 r))
        (liftLzeroto i (RLift a)))) in |- *.
apply oleq_subst; auto.
MCRewrite F0.

ElimAllExistS.
MCRewriteIn F octleq.
MCInv octleq; ElimAllExistS.
MCRewriteIn F1 octleq.
elim (IHTs _ _ _ _ H4 H10); IntrosSplitH.
exists (fun r => SubstL i (x r) (liftLzeroto i (RLift a))).
MCRewrite F'; MCRewrite G; split; try constructor; auto.
intros.
change
  (oleq i
     (SubstL i (omegal_conv (S i) (S i) (refl_equal _) (G0 r))
        (liftLzeroto i (RLift a)))
     (SubstL i (omegal_conv (S i) (S i) (refl_equal _) (x r))
        (liftLzeroto i (RLift a)))) in |- *.
apply oleq_subst.
MCInv octleq; ElimAllExistS.
MCRewrite G0; MCRewrite x.
Qed.



Lemma octeq_instcodety :
 forall T T' Ts G,
 octeq T T' ->
 InstCodeTy T Ts G ->
 exists G' : _, InstCodeTy T' Ts G' /\ octeq (basect G) (basect G').
intros T T' Ts G H; inversion H; intros; eauto.
MCInv InstCodeTy.
exists F'; split; auto; try constructor.
MCRewrite G.

MCInv InstCodeTy.
elim (octleq_listsubstl _ _ _ _ _ H5 H0); IntrosSplitH.
exists (fun r => RUnlift _ (x r) OO); split; auto.
constructor; auto.
apply octleq_octeq; auto.
Qed.


Lemma octeq_wfinstrseq :
 forall CH Is G G',
 octeq (basect G) (basect G') -> WFInstrSeq CH G Is -> WFInstrSeq CH G' Is.
induction Is; intros.
induction i; MCInv WFInstrSeq; repeat MCUnfoldInSelf flook; MCInv octeq.
constructor; try unfold flook in |- *; auto;
 try apply IHIs with (rftupd G r intty); auto.
generalize (H12 r0); MCRewrite (G r0); intros DD; inversion DD; auto.
generalize (H12 r1); MCRewrite (G r1); intros DD; inversion DD; auto.
constructor; intros rr; compare rr r; intros; unfold rftupd in |- *;
 [ MCRewrite rr; replace (beq_reg r r) with true
 | replace (beq_reg r rr) with false ]; auto.

constructor; try unfold flook in |- *; auto;
 try apply IHIs with (rftupd G r intty); auto.
constructor; intros rr; compare rr r; intros; unfold rftupd in |- *;
 [ MCRewrite rr; replace (beq_reg r r) with true
 | replace (beq_reg r rr) with false ]; auto.

apply movl_i with hv T; auto; try unfold flook in |- *; auto;
 try apply IHIs with (rftupd G r (codety T)); auto.
constructor; intros rr; compare rr r; intros; unfold rftupd in |- *;
 [ MCRewrite rr; replace (beq_reg r r) with true
 | replace (beq_reg r rr) with false ]; auto.

MCInv WFInstrSeq.
apply jd_i with G'0 hv T; auto.
apply trans_octeq with (basect G); auto.

MCInv WFInstrSeq; repeat MCUnfoldInSelf flook.
MCInv (octeq (basect G) (basect G')).
generalize (H10 r); MCRewrite (G r); intros DD; inversion DD; auto.
apply jmp_i with G'0 T; try unfold flook in |- *; auto.
apply trans_octeq with (basect G); auto.

elim (octeq_instcodety _ _ _ _ H13 H6); IntrosSplitH.
apply jmp_i with x F'0; try unfold flook in |- *; auto.
apply trans_octeq with (basect G'0); auto.
apply trans_octeq with (basect G); auto.
Qed.



Lemma preserv_jd :
 forall G HT CH H R l Ts H' R' Is' (P:=(H, (R, jd l Ts)))
   (P':=(H', (R', Is'))) (D1 : WFCHeap CH) (D2 : WFHeap H HT)
   (D3 : WFRegFile CH HT R G) (D4 : WFInstrSeq CH G (jd l Ts))
   (D5 : ~ (exists T : _, fmaplook CH l (stub T))) 
   (D6 : Eval CH P P'),
 WFHeap H' HT /\ WFRegFile CH HT R' G /\ WFInstrSeq CH G Is'.
intros.
MCUnfoldIn P (Eval CH P P'); clear P; MCUnfoldIn P' Eval; clear P'.
MCInv Eval.
MCRewrite H'; MCRewrite R'; MCRewrite Is'.
MCRewriteIn Is' fmaplook.
clear D6; MCClear5 H' R' Is' Ts0 l0; MCClear3 R0 H0 CH0.
split; auto.
split; auto.
generalize D1; intros (D1a, D1b).
generalize (D1b _ _ H3); intros.
MCInv WFCHeapVal.
MCInv WFInstrSeq.
UnfoldFmaps; MCInjection (CH l).
MCRewriteIn hv (typeofchv hv = T0).
MCUnfoldIn typeofchv (typeofchv (code G0 Is) = T0).
MCRewriteIn T0 InstCodeTy.
assert (WFInstrSeq CH G' Is).
  apply H4 with Ts; auto.
apply octeq_wfinstrseq with G'; auto.
Qed.






Lemma progress_add :
 forall G HT CH H R rd rs rt Is (P:=(H, (R, iseq (add rd rs rt) Is)))
   (D1 : WFCHeap CH) (D2 : WFHeap H HT) (D3 : WFRegFile CH HT R G)
   (D4 : WFInstrSeq CH G (iseq (add rd rs rt) Is)),
 exists P' : _, Eval CH P P'.
intros.
unfold P in |- *; clear P.
MCInv WFInstrSeq; MCClearAll.
match goal with
| H8:(WFHeap H HT),H1:(WFRegFile _ _ _ _),H2:(flook G rs intty) |- _ =>
    assert (exists s : _, R rs = wi s);
     [ generalize (H1 _ _ H2); intros; MCInv WFWordVal; eauto;
        generalize (wfheap_regheapty H HT H8); intros DH; 
        elim (DH l); auto
     | auto ]
end.
match goal with
| H8:(WFHeap H HT),H1:(WFRegFile _ _ _ _),H2:(flook G rt intty) |- _ =>
    assert (exists t : _, R rt = wi t);
     [ generalize (H1 _ _ H2); intros; MCInv WFWordVal; eauto;
        generalize (wfheap_regheapty H HT H8); intros DH; 
        elim (DH l); auto
     | auto ]
end.
MCElimEx t int; MCElimEx s int.
exists (H, (rfupd R rd (wi (s + t)), Is)).
constructor; auto.
Qed.

Lemma preserv_add :
 forall G HT CH H R rd rs rt Is H' R' Is'
   (P:=(H, (R, iseq (add rd rs rt) Is))) (P':=(H', (R', Is')))
   (D1 : WFCHeap CH) (D2 : WFHeap H HT) (D3 : WFRegFile CH HT R G)
   (D4 : WFInstrSeq CH G (iseq (add rd rs rt) Is)) 
   (D5 : Eval CH P P'),
 exists HT' : _,
   WFHeap H' HT' /\
   WFRegFile CH HT' R' (rftupd G rd intty) /\
   WFInstrSeq CH (rftupd G rd intty) Is'.
intros.
MCUnfoldIn P (Eval CH P P'); clear P; MCUnfoldIn P' Eval; clear P'.
MCInv Eval; MCClear Is0.
(*Unfold i in H5; Unfold i in H6; Unfold i in H7.
MCRewriteIn rs0 '(R rs0)=(wi s); MCRewriteIn rt0 '(R rt0)=(wi t); 
Clear i;*)
 MCClear4 rs0 rt0 R0 H1; MCClear2 CH0 H0.
unfold R'0 in |- *; MCRewrite R0; MCRewrite rd0; MCClear R'0; MCClear2 rd0 R0.
assert (flook G rs intty);
 [ MCInv WFInstrSeq
 | assert (flook G rt intty); [ MCInv WFInstrSeq | auto ] ].
MCRewrite Is'; MCRewrite H'; exists HT; split; auto.
split; auto.

intro; intros.
compare r rd; intros.
MCRewrite r; unfold rfupd in |- *; replace (beq_reg rd rd) with true; auto.
replace t0 with intty; try constructor.
match goal with
| H:(flook (rftupd _ _ _) _ t0) |- _ => generalize H
end.
MCRewrite r; unfold flook, rftupd in |- *; replace (beq_reg rd rd) with true;
 auto; intros DD1; inversion DD1; auto.

unfold rfupd in |- *; replace (beq_reg rd r) with false; auto.
match goal with
| H:(WFRegFile _ _ _ _) |- _ => apply H
end.
match goal with
| H:(flook (rftupd _ _ _) _ _) |- _ => generalize H
end.
unfold flook, rftupd in |- *; replace (beq_reg rd r) with false; auto;
 intros DD1; injection DD1; auto.

MCInv WFInstrSeq.
Qed.

Lemma progress_movi :
 forall G HT CH H R rd i Is (P:=(H, (R, iseq (movi rd i) Is)))
   (D1 : WFCHeap CH) (D2 : WFHeap H HT) (D3 : WFRegFile CH HT R G)
   (D4 : WFInstrSeq CH G (iseq (movi rd i) Is)), exists P' : _, Eval CH P P'.
intros.
unfold P in |- *; clear P.
MCInv WFInstrSeq; MCClearAll.
exists (H, (rfupd R rd (wi i), Is)).
constructor; auto.
Qed.

Lemma preserv_movi :
 forall G HT CH H R rd i Is H' R' Is' (P:=(H, (R, iseq (movi rd i) Is)))
   (P':=(H', (R', Is'))) (D1 : WFCHeap CH) (D2 : WFHeap H HT)
   (D3 : WFRegFile CH HT R G) (D4 : WFInstrSeq CH G (iseq (movi rd i) Is))
   (D5 : Eval CH P P'),
 exists HT' : _,
   WFHeap H' HT' /\
   WFRegFile CH HT' R' (rftupd G rd intty) /\
   WFInstrSeq CH (rftupd G rd intty) Is'.
intros.
MCUnfoldIn P (Eval CH P P'); clear P; MCUnfoldIn P' Eval; clear P'.
MCInv Eval; MCClear Is0.
(*Unfold i0 in H5; Unfold i0 in H6.
Clear i0;*) MCClear2 R0 H1; MCClear2 CH0 H0.
unfold R'0 in |- *; MCRewrite R0; MCRewrite rd0; MCClear R'0; MCClear2 rd0 R0.
MCRewrite Is'; MCRewrite H'; exists HT; split; auto.
split; auto.

intro; intros.
compare r rd; intros.
MCRewrite r; unfold rfupd in |- *; replace (beq_reg rd rd) with true; auto.
replace t with intty; try constructor.
match goal with
| H:(flook (rftupd _ _ _) _ t) |- _ => generalize H
end.
MCRewrite r; unfold flook, rftupd in |- *; replace (beq_reg rd rd) with true;
 auto; intros DD1; inversion DD1; auto.

unfold rfupd in |- *; replace (beq_reg rd r) with false; auto.
match goal with
| H:(WFRegFile _ _ _ _) |- _ => apply H
end.
match goal with
| H:(flook (rftupd _ _ _) _ _) |- _ => generalize H
end.
unfold flook, rftupd in |- *; replace (beq_reg rd r) with false; auto;
 intros DD1; injection DD1; auto.

MCInv WFInstrSeq.
Qed.


Lemma progress_movl :
 forall G HT CH H R rd l Is (P:=(H, (R, iseq (movl rd l) Is)))
   (D1 : WFCHeap CH) (D2 : WFHeap H HT) (D3 : WFRegFile CH HT R G)
   (D4 : WFInstrSeq CH G (iseq (movl rd l) Is)), exists P' : _, Eval CH P P'.
intros.
unfold P in |- *; clear P.
MCInv WFInstrSeq.
exists (H, (rfupd R rd (wl l), Is)).
constructor; auto.
Qed.

Lemma preserv_movl :
 forall G HT CH H R rd l Is H' R' Is' (P:=(H, (R, iseq (movl rd l) Is)))
   (P':=(H', (R', Is'))) (D1 : WFCHeap CH) (D2 : WFHeap H HT)
   (D3 : WFRegFile CH HT R G) (D4 : WFInstrSeq CH G (iseq (movl rd l) Is))
   (D5 : Eval CH P P') hv T' (D6 : fmaplook CH l hv) 
   (D7 : typeofchv hv = T'),
 exists HT' : _,
   WFHeap H' HT' /\
   WFRegFile CH HT' R' (rftupd G rd (codety T')) /\
   WFInstrSeq CH (rftupd G rd (codety T')) Is'.
intros.
MCUnfoldIn P (Eval CH P P'); clear P; MCUnfoldIn P' Eval; clear P'.
MCInv Eval; MCClear Is0.
(*Unfold i in H5; Unfold i in H6.
Clear i; MCClear4 rs0 rt0*) MCClear2 R0 H1; MCClear2 CH0 H0.
unfold R'0 in |- *; MCRewrite R0; MCRewrite rd0; MCClear R'0; MCClear2 rd0 R0.
inversion D4.
MCRewrite Is'; MCRewrite H'; exists HT; split; auto.
replace T' with T.
split; auto.

MCRewrite l0; intro; intros.
compare r rd; intros.
MCRewrite r; unfold rfupd in |- *; replace (beq_reg rd rd) with true; auto.
replace t with (codety T).

induction  hv0 as [c i| c].
MCUnfoldIn typeofchv (typeofchv (code c i) = T); MCRewrite T.
apply label_wval0 with c i; auto.
MCUnfoldIn typeofchv (typeofchv (stub c) = T); MCRewrite T.
apply label_wval1 with c; auto.

match goal with
| H:(flook (rftupd _ _ _) _ t) |- _ => generalize H
end.
MCRewrite r; unfold flook, rftupd in |- *; replace (beq_reg rd rd) with true;
 auto; intros DD1; injection DD1; auto.

unfold rfupd in |- *; replace (beq_reg rd r) with false; auto.
match goal with
| H:(WFRegFile _ _ _ _) |- _ => apply H
end.
match goal with
| H:(flook (rftupd _ _ _) _ _) |- _ => generalize H
end.
unfold flook, rftupd in |- *; replace (beq_reg rd r) with false; auto;
 intros DD1; injection DD1; auto.

UnfoldFmaps; MCInjection (CH l); intros.
rewrite H11 in D7.
MCRewrite T'.
Qed.



