(* 
 *
 * Coq code for the paper
 * 
 *      Interfacing Hoare Logic and Type Systems for FPCC
 *         Nadeem Abdul Hamid, 
 *
 * (For Coq version 8.0)
 *
 *)

Require Import EqNat.
Require Import List.
Require Import mystuff.
Require Import natutil.
Require Import tis.

(***************************************************************************)
(* Definitions *)

Definition Addr := Word.
Definition Pred := State -> Prop.
Definition CdSpec := Word -> optT (prodT nat Pred). (* cmds length & precondtn *)
Definition CmdList := list Cmd.
Definition WordList := list Word.

Fixpoint Flatten (Cs : CmdList) (M : Mem) (pc : Addr) {struct Cs} : Prop :=
  match Cs with
  | nil => True
  | c :: Cs' => Dc (M pc) = c /\ Flatten Cs' M (S pc)
  end.

Definition IsCodeArea (Ct : CdSpec) (w : Addr) (n : Word) : Prop :=
  exists P : _,
    (exists f : _,
       (exists s : _,
          (exists m : _,
             Ct f = someT (pairT s P) /\ w + n <= f + s /\ w = f + m))).

Definition curmem (St : State) := let (M, X) := St in M.
Definition currf (St : State) := let (M, X) := St in let (R, pc) := X in R.
Definition curpc (St : State) := let (M, X) := St in let (R, pc) := X in pc.
Definition curcmd (St : State) (c : Cmd) := Dc (curmem St (curpc St)) = c.

(***************************************************************************)
(* "typing" rules *)

Section CAPRules.

Variable SP : Pred. 


Inductive WFCapCmds : CdSpec -> Pred -> CmdList -> Prop :=
  | wfcapcmd :
      forall (Ct : CdSpec) (P : Pred) c Cs rd rs rt i,
      c = _add rd rs rt \/
      c = _addi rd rs i \/
      c = _sub rd rs rt \/
      c = _subi rd rs i \/
      c = _mov rd rs \/ c = _movi rd i \/ c = _ld rd rs i ->
      (forall St : State, P St -> curcmd St c -> SP St) ->
      (forall St : State,
       P St ->
       curcmd St c -> exists Q : Pred, Q (Step St) /\ WFCapCmds Ct Q Cs) ->
      WFCapCmds Ct P (c :: Cs)
  | wfcapst :
      forall (Ct : CdSpec) (P : Pred) Cs rd rs i,
      let c := _st rd i rs in
      (forall St : State,
       P St -> curcmd St c -> ~ IsCodeArea Ct (currf St rd + i) 1) ->
      (forall St : State, P St -> curcmd St c -> SP St) ->
      (forall St : State,
       P St ->
       curcmd St c -> exists Q : Pred, Q (Step St) /\ WFCapCmds Ct Q Cs) ->
      WFCapCmds Ct P (c :: Cs)
  | wfcapbgt :
      forall (Ct : CdSpec) (P : Pred) Cs rs rt f,
      let c := _bgt rs rt f in
      (forall St : State, P St -> curcmd St c -> SP St) ->
      (forall St : State,
       P St ->
       curcmd St c ->
       currf St rs > currf St rt ->
       exists n : _, (exists R : _, Ct f = someT (pairT n R) /\ R (Step St))) ->
      (forall St : State,
       P St ->
       curcmd St c ->
       currf St rs <= currf St rt ->
       exists Q : Pred, Q (Step St) /\ WFCapCmds Ct Q Cs) ->
      WFCapCmds Ct P (c :: Cs)
  | wfcapbgti :
      forall (Ct : CdSpec) (P : Pred) Cs rs i f,
      let c := _bgti rs i f in
      (forall St : State, P St -> curcmd St c -> SP St) ->
      (forall St : State,
       P St ->
       curcmd St c ->
       currf St rs <= i -> exists Q : Pred, Q (Step St) /\ WFCapCmds Ct Q Cs) ->
      (forall St : State,
       P St ->
       curcmd St c ->
       currf St rs > i ->
       exists n : _, (exists R : _, Ct f = someT (pairT n R) /\ R (Step St))) ->
      WFCapCmds Ct P (c :: Cs)
  | wfcapjd :
      forall (Ct : CdSpec) (P : Pred) f,
      (forall St : State, P St -> curcmd St (_jd f) -> SP St) ->
      (forall St : State,
       P St ->
       curcmd St (_jd f) ->
       exists n : _, (exists Q : _, Ct f = someT (pairT n Q) /\ Q (Step St))) ->
      WFCapCmds Ct P (_jd f :: nil)
  | wfcapjmp :
      forall (Ct : CdSpec) (P : Pred) r,
      (forall St : State, P St -> curcmd St (_jmp r) -> SP St) ->
      (forall St : State,
       P St ->
       curcmd St (_jmp r) ->
       exists n : _,
         (exists Q : _, Ct (currf St r) = someT (pairT n Q) /\ Q (Step St))) ->
      WFCapCmds Ct P (_jmp r :: nil).


Definition WFCapCdSpec (M : Mem) (Ct : CdSpec) :
  Prop (*    |-sp  M : Ct  *)
   :=
  forall f n P,
  Ct f = someT (pairT n P) ->
  exists Cs : _, Flatten Cs M f /\ length Cs = n /\ WFCapCmds Ct P Cs.


Inductive WFCapState : State -> Prop :=
    wfcapstate :
      forall M R pc Ct Cs P,
      WFCapCdSpec M Ct -> (*        |-sp   M : Ct          *)
      Flatten Cs M pc -> (*        |- Flatten Cs => M@pc  *)
      WFCapCmds Ct P Cs -> (*  Ct; P |-sp   Cs              *)
      P (M, (R, pc)) -> (*       P (M,R,pc)              *)
      IsCodeArea Ct pc (length Cs) -> 
      (*   IsCodeArea(Ct,pc)         *)
      (*  ___________________________  *)
      WFCapState (M, (R, pc)). (*        |-sp  (M,R,pc)         *)




(***************************************************************************)
(* Properties of CAP system *)



Lemma IsCodeArea_n_Sn :
 forall Ct f n, IsCodeArea Ct f (S n) -> IsCodeArea Ct (S f) n.
unfold IsCodeArea in |- *; intros Ct f n.
intros (P, (f', (s, (m, H)))); IntrosSplitH.
Exists4 P f' s (S m); intros.
repeat split; auto.
replace (S f + n) with (S (f + n)); auto.
replace (S (f + n)) with (f + S n); auto.
replace (f' + S m) with (S (f' + m)); auto.
Qed.



Lemma le_Sm_1 : forall f x l, S (f + x) <= l -> S f <= l.
induction f; UnfoldFold plus; intros; auto.
MCInv le.
apply le_n_S; apply le_O_n.
apply le_n_S; apply le_O_n.

MCInv le.
apply le_n_S; apply IHf with x; auto.
apply le_n_S; apply IHf with x; auto.
apply le_Sn_le; auto.
Qed.



Lemma Flatten_noncodeupd :
 forall M Ct Cs f x w,
 (*    (WFCapCdSpec M Ct) -> *)
 Flatten Cs M f ->
 IsCodeArea Ct f (length Cs) ->
 ~ IsCodeArea Ct x 1 -> Flatten Cs (updatemem M x w) f.
induction Cs; UnfoldFold Flatten; IntrosSplitH; auto.
cut (x <> f); intros.
split; auto.
unfold updatemem in |- *; replace (beq_nat f x) with false; auto.
apply beq_neq_false; auto.

apply IHCs; auto.
apply IsCodeArea_n_Sn; auto.


intro.
cut (IsCodeArea Ct f 1); intros.
MCRewriteIn x (~ IsCodeArea Ct x 1).
MCElimEx4 P f' m l (IsCodeArea Ct f (length (a :: Cs))).
Exists4 P f' m l; intros.
repeat split; auto.
replace (f + 1) with (S f); auto.
apply le_Sm_1 with (length Cs).
replace (S (f + length Cs)) with (f + S (length Cs)); auto.
replace (f + 1) with (S (f + 0)); auto.
Qed.


Lemma WFCapCdSpec_noncodeupd :
 forall M Ct x w,
 WFCapCdSpec M Ct -> ~ IsCodeArea Ct x 1 -> WFCapCdSpec (updatemem M x w) Ct.
unfold WFCapCdSpec in |- *; intros.
elim (H _ _ _ H1); intros Cs; IntrosSplitH.
exists Cs.
split; [ auto | split; auto ].
apply Flatten_noncodeupd with Ct; auto.
MCRewriteRev n.
Exists4 P f n 0.
repeat split; auto.
Qed.



Lemma CapPreserv : forall St, WFCapState St -> WFCapState (Step St).
intros.
MCNewInd WFCapState; MCNewInd WFCapCmds.

repeat
 match goal with
 | HH:(_ \/ _) |- _ =>
     MCInv or; clear HH;
      [ MCRewriteIn c (Flatten (c :: Cs) M pc);
         MCRewriteIn c
          (forall St : State,
           P St ->
           curcmd St c -> exists Q : Pred, Q (Step St) /\ WFCapCmds Ct Q Cs);
         MCRewriteIn c (forall St : State, P St -> curcmd St c -> SP St);
         MCInv Flatten;
         match goal with
         | HH:(forall St : State, P St -> curcmd St _ -> exists Q : _, _) |-
         _ => generalize (HH (M, (R, pc)))
         end; unfold Step in |- *; MCRewrite (Dc (M pc)); 
         intros HH2; elim HH2;
         [ intros Q; IntrosSplitH; apply wfcapstate with Ct Cs Q; auto;
            apply IsCodeArea_n_Sn; auto
         | auto
         | auto ]
      | auto ]
 end.

(* last case *)
MCRewriteIn c (Flatten (c :: Cs) M pc);
 MCRewriteIn c
  (forall St : State,
   P St -> curcmd St c -> exists Q : Pred, Q (Step St) /\ WFCapCmds Ct Q Cs);
 MCRewriteIn c (forall St : State, P St -> curcmd St c -> SP St);
 MCInv Flatten;
 match goal with
 | HH:(forall St : State, P St -> curcmd St _ -> exists Q : _, _) |- _ =>
     generalize (HH (M, (R, pc)))
 end; unfold Step in |- *; MCRewrite (Dc (M pc)); intros HH2; 
 elim HH2;
 [ intros Q; IntrosSplitH; apply wfcapstate with Ct Cs Q; auto;
    apply IsCodeArea_n_Sn; auto
 | auto
 | auto ].

(* st case *)
MCUnfoldIn c (Flatten (c :: Cs) M pc);
 MCUnfoldIn c
  (forall St : State,
   P St -> curcmd St c -> exists Q : Pred, Q (Step St) /\ WFCapCmds Ct Q Cs);
 MCUnfoldIn c (forall St : State, P St -> curcmd St c -> SP St);
 MCInv Flatten.
  match goal with
  | HH:(forall St : State, P St -> curcmd St _ -> exists Q : _, _) |- _ =>
      generalize (HH (M, (R, pc)))
  end; unfold Step in |- *; MCRewrite (Dc (M pc)); 
   intros HH2; elim HH2; [ intros Q; IntrosSplitH | auto | auto ].
  apply wfcapstate with Ct Cs Q; auto; fold Flatten in H7.
apply WFCapCdSpec_noncodeupd; auto.
match goal with
| HH:(forall St : State, P St -> curcmd St _ -> ~ IsCodeArea _ _ _) |- _ =>
    generalize (HH (M, (R, pc))); unfold currf in |- *; intros; auto
end.
apply Flatten_noncodeupd with Ct; auto.
apply IsCodeArea_n_Sn; replace (S (length Cs)) with (length (c :: Cs));
 try compute in |- *; auto.
match goal with
| HH:(forall St : State, P St -> curcmd St _ -> ~ IsCodeArea _ _ _) |- _ =>
    generalize (HH (M, (R, pc))); unfold currf in |- *; intros; auto
end.
apply IsCodeArea_n_Sn; replace (S (length Cs)) with (length (c :: Cs));
 try compute in |- *; auto.

(* bgt case *)
unfold c in H1; unfold c in H4; unfold c in H5.
MCInv Flatten.
fold Flatten in H7.
elim (le_or_gt (R rt) (R rs)); intros.
 generalize (H5 (M, (R, pc))); unfold Step in |- *; MCRewrite (Dc (M pc));
  unfold c in |- *; replace (blt_nat (R rt) (R rs)) with false;
  try symmetry  in |- *; try apply le_blt_false; auto; 
  intros HH2; elim HH2; auto.
 intros Q; IntrosSplitH.
 apply wfcapstate with Ct Cs Q; auto; apply IsCodeArea_n_Sn;
  replace (S (length Cs)) with (length (c :: Cs)); 
  try compute in |- *; auto.
generalize (H4 (M, (R, pc)) H2 H6 H8); unfold Step in |- *;
 MCRewrite (Dc (M pc)); unfold c in |- *;
 replace (blt_nat (R rt) (R rs)) with true; try symmetry  in |- *;
 try apply blt_lt_true; auto.
 intros (n', (P', HH)); SplitAllHyps.
 elim (H (*WFCapCdSpec*)  _ _ _ H9); IntrosSplitH.
 apply wfcapstate with Ct x P'; auto.
 MCRewriteRev n'; Exists4 P' f n' 0; repeat split; auto.

(* bgti case *)
unfold c in H1; unfold c in H4; unfold c in H5.
MCInv Flatten.
fold Flatten in H7.
elim (le_or_gt i (R rs)); intros.
 generalize (H4 (M, (R, pc))); unfold Step in |- *; MCRewrite (Dc (M pc));
  unfold c in |- *; replace (blt_nat i (R rs)) with false;
  try symmetry  in |- *; try apply le_blt_false; auto; 
  intros HH2; elim HH2; auto.
   intros Q; IntrosSplitH.
 apply wfcapstate with Ct Cs Q; auto; apply IsCodeArea_n_Sn;
  replace (S (length Cs)) with (length (c :: Cs)); 
  try compute in |- *; auto.
generalize (H5 (M, (R, pc)) H2 H6 H8); unfold Step in |- *;
 MCRewrite (Dc (M pc)); unfold c in |- *;
 replace (blt_nat i (R rs)) with true; try symmetry  in |- *;
 try apply blt_lt_true; auto.
 intros (n', (P', HH)); SplitAllHyps.
 elim (H (*WFCapCdSpec*)  _ _ _ H9); IntrosSplitH.
 apply wfcapstate with Ct x P'; auto.
 MCRewriteRev n'; Exists4 P' f n' 0; repeat split; auto.

(* jd *)
MCInv Flatten.
generalize (H4 (M, (R, pc)) H2 H5); unfold Step in |- *;
 MCRewrite (Dc (M pc)).
intros (n', (P', HH)); SplitAllHyps.
elim (H _ _ _ H7); IntrosSplitH.
apply wfcapstate with Ct x P'; auto.
MCRewriteRev n'; Exists4 P' f n' 0; repeat split; auto.

(* jmp *)
MCInv Flatten.
generalize (H4 (M, (R, pc)) H2 H5); unfold Step in |- *;
 MCRewrite (Dc (M pc)).
intros (n', (P', HH)); SplitAllHyps.
elim (H _ _ _ H7); IntrosSplitH.
apply wfcapstate with Ct x P'; auto.
MCRewriteRev n'; Exists4 P' (R r) n' 0; repeat split; auto.
Qed.


Lemma WFCapState_SP : forall St, WFCapState St -> SP St.
intros.
MCInv WFCapState.
MCInv WFCapCmds; MCRewriteIn Cs (Flatten Cs M pc); MCInv Flatten;
 match goal with
 | HH:(forall St : State, P St -> curcmd _ _ -> SP St) |- _ =>
     apply HH; auto
 end.
Qed.


End CAPRules.





(***************************************************************************)
(* Definitions for FPCC package production *)

(* Definition of safety *)
Definition Safe (St : State) (SP : Pred) := forall n, SP (MultiStep n St).

(* The most basic safety policy *)
Definition SPbase (St : State) := ~ curcmd St _ill.

Lemma WFCap_SPbase : forall SP St, WFCapState SP St -> SPbase St.
unfold SPbase in |- *; intros.
MCInv WFCapState.
MCInv WFCapCmds; MCRewriteIn Cs (Flatten Cs M pc); MCInv Flatten;
 unfold curcmd, curmem, curpc in |- *; MCRewrite (Dc (M pc));
 try unfold c in |- *; intro Hbad; try discriminate Hbad.
absurd (c = _ill); auto; clear Hbad.
MCInv or; [ MCRewrite c; intro Hbadd; discriminate Hbadd | auto ].
MCInv or; [ MCRewrite c; intro Hbadd; discriminate Hbadd | auto ].
MCInv or; [ MCRewrite c; intro Hbadd; discriminate Hbadd | auto ].
MCInv or; [ MCRewrite c; intro Hbadd; discriminate Hbadd | auto ].
MCInv or; [ MCRewrite c; intro Hbadd; discriminate Hbadd | auto ].
MCInv or; [ MCRewrite c; intro Hbadd; discriminate Hbadd | auto ].
MCRewrite c; intro Hbadd; discriminate Hbadd.
Qed.

Lemma WFCap_SPadd :
 forall (PA SPadd : Pred) St,
 WFCapState PA St -> (forall St, PA St -> SPadd St) -> SPbase St /\ SPadd St.
intros.
split; try apply WFCap_SPbase with PA; auto.
apply H0; apply WFCapState_SP; auto.
Qed.

Lemma CapPreservMulti :
 forall n (SP : Pred) St, WFCapState SP St -> WFCapState SP (MultiStep n St).
induction n; auto.
intros.
UnfoldFold MultiStep.
apply CapPreserv; auto.
Qed.

Theorem WFCap_Safe :
 forall (PA SPadd : Pred) St,
 WFCapState PA St ->
 (forall St, PA St -> SPadd St) -> Safe St (fun S => SPbase S /\ SPadd S).
unfold Safe in |- *; intros.
induction n.
unfold MultiStep in |- *.
apply WFCap_SPadd with PA; auto.
apply WFCap_SPadd with PA; auto.
apply CapPreservMulti; auto.
Qed.
