 (* 
 * Coq code for the paper
 * 
 *      Interfacing Hoare Logic and Type Systems for FPCC
 *         Nadeem Abdul Hamid, 
 *
 * (For Coq version 8.0)
 *
 *)

(* Generic definition of partial functions and reasoning utilities for
   them (e.g. separation logic stuff) *)

Require Import Arith.
Require Import Compare.
Require Import EqNat.
Require Export Decidable.
Require Export List.
Require Import Le.
Require Import Lt.
Require Import mystuff.
Require Import natutil.


(***************************************************************************)
(* e.g.  (listnth (cons a (cons b (cons c nil))) 1) ==> (Some ? b) *)
Definition listnth : forall A : Set, list A -> nat -> option A := nth_error.



(***************************************************************************)
(***************************************************************************)
(***************************************************************************)
Section FMAP.
Set Implicit Arguments.
Unset Strict Implicit.

Variables A B : Set.
Variable aeq_dec : forall a a' : A, a = a' \/ a <> a'.
Variable beq_aset : A -> A -> bool.
Variable aneq_beq_false : forall a a', a <> a' -> false = beq_aset a a'.



(***************************************************************************)
(* Basic definitions of partial functions and predicates thereupon *)
Definition fmap := A -> option B.
Definition fmapPred := fmap -> Prop.

Definition emptyfmap : fmap := fun _ => None.

(***************************************************************************)
(* lookup, update, and delete *)
Definition fmaplook (M : fmap) a b := M a = Some b.

Definition fmapupd (M : fmap) a b a' := if beq_aset a' a then b else M a'.

Definition fmapdel (M : fmap) a a' := if beq_aset a' a then None else M a'.



(***************************************************************************)
(* reasoning about domains *)
Definition indomf (M : fmap) a : Prop := exists b : _, fmaplook M a b.
Definition notindomf (M : fmap) a : Prop := M a = None.
Definition nulldomf (M : fmap) : Prop := forall a, notindomf M a.

Definition onetoonef (M : fmap) : Prop :=
  forall a a' b, fmaplook M a b -> fmaplook M a' b -> a = a'.

Lemma not_indom_notindom : forall M a, ~ indomf M a -> notindomf M a.
unfold indomf, notindomf in |- *; intros.
assert (exists W : _, M a = W); eauto; intros.
ElimEx H0 W; induction W; auto.
MCUnfoldIn fmaplook not; MCRewriteIn (M a) not.
elim H; eauto.
Qed.

Lemma notindom_not_indom : forall M a, notindomf M a -> ~ indomf M a.
unfold indomf, notindomf in |- *; intros; intro.
ElimEx H0 w; MCUnfoldIn fmaplook fmaplook; MCRewriteIn (M a) (M a = None).
discriminate H.
Qed.

Lemma indomf_dec : forall M a, decidable (indomf M a).
intros; unfold indomf in |- *.
unfold fmaplook in |- *; elim (M a); intros; unfold decidable in |- *; eauto.
right.
intro.
MCInv ex; MCInv eq.
Qed.

Lemma fmapdel_onetoone : forall M a, onetoonef M -> onetoonef (fmapdel M a).
unfold onetoonef, fmaplook, fmapdel in |- *; intros M a H a1 a2 b.
elim (beq_aset a1 a); intros H1; try AbsurdEq.
elim (beq_aset a2 a); intros; try AbsurdEq.
apply H with b; auto.
Qed.




(***************************************************************************)
(* conversion to list form *)
Fixpoint fmap_list (Ls : list (A * B)) (M : fmap) {struct Ls} : Prop :=
  match Ls with
  | nil => nulldomf M
  | (a, b) :: Ls' => fmaplook M a b /\ fmap_list Ls' (fmapdel M a)
  end.

(*
Definition fmap_list [M:fmap; Ls:(list (A*B))] : Prop
  := ((a:?)(indomf M a) 
            <-> (EX n | (EX b | (listnth ? Ls n)=(Some ? (a,b))
                    /\ ((m,a',b':?)(listnth ? (tail Ls) m)=(Some ? (a',b')) ->
	                              ~a=a' )))).
*)

(***************************************************************************)
(* about fmap_list *)
Lemma fmap_listnth_fmaplook :
 forall Ls M a b,
 fmap_list Ls M ->
 fmaplook M a b -> exists n : _, listnth _ Ls n = Some (a, b).
induction Ls; UnfoldFold fmap_list.
unfold nulldomf, fmaplook, notindomf in |- *; intros.
generalize (H a); intros; AbsurdEq.
intros; induction  a as (a, b0); SplitAllHyps.
elim (aeq_dec a0 a); intros ee.
rewrite ee; exists 0; compute in |- *; auto.
repeat MCUnfoldIn fmaplook fmaplook.
MCRewriteIn a (M a = Some b0); MCInjection (M a0); MCRewrite b; auto.

assert (fmaplook (fmapdel M a) a0 b).
   generalize (aneq_beq_false ee); intros.
   unfold fmapdel, fmaplook in |- *; MCRewrite (beq_aset a0 a).
elim (IHLs _ _ _ H2 H); intros.
exists (S x).
unfold listnth, nth_error in |- *; auto.
Qed.




(***************************************************************************)
(* equivalence, union, disjuction *)
Definition eqf (M M' : fmap) : Prop := forall l, M l = M' l.

Definition disjf (M M' : fmap) : Prop :=
  forall l, ~ (indomf M l /\ indomf M' l).

Definition appendf (M M' : fmap) (_ : disjf M M') : fmap :=
  fun a => match M a with
           | Some n => M a
           | _ => M' a
           end.




(***************************************************************************)
(* predicates *)
Definition truefp : fmapPred := fun _ => True.

Definition singfp a w : fmapPred :=
  fun M => fmaplook M a w /\ (forall a', a <> a' -> notindomf M a').

Definition starfp (P1 P2 : fmapPred) : fmapPred :=
  fun M =>
  exists M1 : _,
    (exists M2 : _,
       (exists D : disjf M1 M2,
          P1 M1 /\ P2 M2 /\ eqf (appendf (M:=M1) (M':=M2) D) M)).

Ltac ElimStar H Mp1 Mp2 := ElimEx2 H Mp1 Mp2; elim H; clear H; IntrosSplitH.



(***************************************************************************)
(* reasoning about starfp, eqf, disjf, *)
Lemma eqf_append_eq_left :
 forall Mp Mp1 Mp2 a b (D : disjf Mp1 Mp2),
 eqf (appendf D) Mp -> fmaplook Mp1 a b -> fmaplook Mp a b.
unfold eqf, appendf, fmaplook, disjf in |- *; intros.
generalize (H a); MCRewrite (Mp1 a); intros; auto.
Qed.

Lemma eqf_refl : forall Mp, eqf Mp Mp.
unfold eqf in |- *; auto.
Qed.
Hint Immediate eqf_refl.

Lemma indomf_append_left :
 forall Mp1 Mp2 l (D : disjf Mp1 Mp2), indomf Mp1 l -> indomf (appendf D) l.
unfold disjf, indomf, appendf, fmaplook in |- *; intros.
ElimExIn H b; MCRewrite (Mp1 l); auto.
Qed.


Lemma indomf_append_right :
 forall Mp1 Mp2 l (D : disjf Mp1 Mp2), indomf Mp2 l -> indomf (appendf D) l.
unfold disjf, indomf, appendf, fmaplook in |- *; intros.
elim (Mp1 l); eauto.
Qed.


Lemma disjf_append_left :
 forall Mp Mp' Mp1 Mp2 (D1 : disjf Mp1 Mp2),
 eqf Mp' (appendf (M:=Mp1) (M':=Mp2) D1) -> disjf Mp Mp' -> disjf Mp Mp1.
unfold disjf in |- *; intros.
intro; SplitAllHyps.
elim (indomf_dec Mp2 l); intros.
elim (D1 l); split; auto.
generalize (indomf_append_left D1 H3); intros.
elim (H0 l); split; auto.
generalize (H l); intros; unfold indomf, fmaplook in |- *; MCRewrite (Mp' l).
Qed.


Lemma disjf_sym : forall Mp1 Mp2, disjf Mp1 Mp2 -> disjf Mp2 Mp1.
unfold disjf, not in |- *; IntrosSplitH.
apply H with l; split; auto.
Qed.

Lemma appendf_left_indom :
 forall Mp Mp1 Mp2 l (D : disjf Mp1 Mp2),
 eqf (appendf D) Mp -> indomf Mp1 l -> indomf Mp l.
unfold eqf, indomf, appendf, fmaplook in |- *; intros.
ElimEx H0 b.
generalize (H l); MCRewrite (Mp1 l); intros; eauto.
Qed.

Lemma eqf_append_disj_left :
 forall Mp Mp' Mp1 Mp2 (D1 : disjf Mp Mp') (D2 : disjf Mp1 Mp2),
 eqf (appendf D2) Mp' -> disjf Mp Mp1.
unfold disjf, eqf in |- *; intros; intro; SplitAllHyps.
apply D1 with l; split; auto.
apply appendf_left_indom with Mp1 Mp2 D2; auto.
Qed.


Lemma indomf_append_or :
 forall Mp1 Mp2 l (D : disjf Mp1 Mp2),
 indomf (appendf D) l -> indomf Mp1 l \/ indomf Mp2 l.
unfold appendf, indomf, fmaplook in |- *; intros Mp1 Mp2 l; elim (Mp1 l);
 intros; eauto.
Qed.

Lemma appendf_left_disj :
 forall Mp Mp1 Mp2 (D1 : disjf Mp Mp1) (D2 : disjf Mp1 Mp2),
 disjf Mp (appendf (M:=Mp1) (M':=Mp2) D2) ->
 disjf (appendf (M:=Mp) (M':=Mp1) D1) Mp2.
unfold disjf in |- *; intros; intro; SplitAllHyps.
elim (indomf_append_or H1); intros; auto.
apply (H l); split; auto.
apply indomf_append_right; auto.
apply (D2 l); split; auto.
Qed.

Lemma eqf_sym : forall Mp Mp', eqf Mp Mp' -> eqf Mp' Mp.
unfold eqf in |- *; intros; auto.
Qed.


Lemma disjf_eq_disj :
 forall Mp Mp' Mp'', disjf Mp Mp' -> eqf Mp' Mp'' -> disjf Mp Mp''.
unfold disjf, eqf, indomf, fmaplook in |- *; intros; intro; SplitAllHyps.
apply H with l; split; auto.
replace (Mp' l) with (Mp'' l); try symmetry  in |- *; try apply H0; auto.
Qed.


Lemma eqf_subst_right :
 forall M Mp MpB Mp1 Mp2 D1 D2,
 eqf (appendf (M:=Mp) (M':=MpB) D1) M ->
 eqf (appendf (M:=Mp1) (M':=Mp2) D2) MpB ->
 exists D3 : _,
   eqf (appendf (M:=Mp) (M':=appendf (M:=Mp1) (M':=Mp2) D2) D3) M.
unfold eqf in |- *; intros.
assert (disjf Mp (appendf D2)).
   apply disjf_eq_disj with MpB; auto.
   unfold eqf in |- *; intros; symmetry  in |- *; auto.
exists H1; intros.
generalize (H l); intros.
MCRewrite (M l); unfold appendf in |- *.
elim (Mp l); auto.
generalize (H0 l); intros; MCRewrite (MpB l).
Qed.


Lemma eqf_assoc_left :
 forall M Mp Mp1 Mp2 D1 D2,
 eqf (appendf (M:=Mp) (M':=appendf (M:=Mp1) (M':=Mp2) D1) D2) M ->
 exists D3 : _,
   (exists D4 : _,
      eqf (appendf (M:=appendf (M:=Mp) (M':=Mp1) D3) (M':=Mp2) D4) M).
intros; unfold eqf in |- *.
assert (eqf (appendf D1) (appendf D1)); auto.
generalize (disjf_append_left H0 D2); intros D3; exists D3.
generalize
 (appendf_left_disj (Mp:=Mp) (Mp1:=Mp1) (Mp2:=Mp2) (D1:=D3) (D2:=D1) D2);
 intros D4; exists D4.
intros; generalize (H l); unfold appendf in |- *; elim (Mp l); intros; auto.
Qed.




Lemma star_assoc_left :
 forall P P1 P2 (M : fmap),
 starfp P (starfp P1 P2) M -> starfp (starfp P P1) P2 M.
intros.
ElimStar H Mp MpB.
ElimStar H Mp1 Mp2.

elim (eqf_subst_right H2 H4); intros.
assert (disjf Mp Mp1).
   apply eqf_append_disj_left with (appendf x0) Mp2 x0; auto.
match goal with
| H5:(disjf Mp Mp1) |- _ => Exists2 (appendf H5) Mp2
end.
generalize
 (appendf_left_disj (Mp:=Mp) (Mp1:=Mp1) (Mp2:=Mp2) (D1:=H5) (D2:=x0) x1);
 intros Mpapdisj; exists Mpapdisj.
split; [ auto | split; auto ].
Exists3 Mp Mp1 H5; split; [ auto | split; auto ].

generalize H3; unfold appendf, eqf in |- *; intros.
generalize (H6 l); elim (Mp l); intros; auto.
Qed.


(***************************************************************************)
(* a partial function is the same as a total function over its domain *)
Definition satisfmapfull (MM : A -> B) (M : fmap) : Prop :=
  forall a b, fmaplook M a b -> MM a = b.





(***************************************************************************)
Set Strict Implicit.
Unset Implicit Arguments.

End FMAP.
(***************************************************************************)
(***************************************************************************)
(***************************************************************************)




(***************************************************************************)
(* Tactics *)
Ltac UnfoldFmaps :=
  repeat
   match goal with
   | H:(fmaplook _ _ _) |- _ => unfold fmaplook in H
   |  |- (fmaplook _ _ _) => unfold fmaplook in |- *
   end.

Ltac ElimStar H Mp1 Mp2 := ElimEx2 H Mp1 Mp2; elim H; clear H; IntrosSplitH.

Hint Immediate eqf_refl.


(***************************************************************************)
(***************************************************************************)
Definition coversfnat B (M : fmap nat B) min max : Prop :=
  forall a, indomf M a <-> min <= a /\ a < max.


Definition eqdomf A B C (M : fmap A B) (M' : fmap A C) : Prop :=
  forall a, indomf M a <-> indomf M' a.

Definition fmapdelN (B : Set) := fmapdel (A:=nat) (B:=B) beq_nat.
Definition fmap_listN (B : Set) := fmap_list (A:=nat) (B:=B) beq_nat.


(* lemmas about fmapdelN *)
Lemma fmapdelN_neq_indomf :
 forall B M a a0, indomf (fmapdelN B M a) a0 -> a0 <> a -> indomf M a0.
unfold fmapdelN, fmapdel, indomf, fmaplook in |- *; intros.
generalize (beq_neq_false a0 a H0); intros ee.
rewrite <- ee in H; auto.
Qed.




(* Lemmas about fmap_list when dealing with integer domains *)
Definition fstP : forall A B, A /\ B -> A.
intros; tauto. Defined.
Definition sndP : forall A B, A /\ B -> B.
intros; tauto. Defined.



Lemma listnth_nil_false : forall A n, listnth A nil n = None.
induction n; compute in |- *; auto.
Qed.
Hint Immediate listnth_nil_false.

(*
Lemma unique_keylist_nil : (A,B:?)(!unique_keylist A B (nil ?)).
Intros; Intro; Intros.
Induction m; MCUnfoldIn listnth eq; MCUnfoldIn nth_error eq.
MCUnfoldIn error eq; AbsurdEq.
MCUnfoldIn error eq; AbsurdEq.
Save.
Hints Immediate unique_keylist_nil listnth_nil_false.
*)


Lemma fmapdelN_commut_eq :
 forall (B : Set) M a a0,
 eqf (fmapdelN B (fmapdelN _ M a0) a) (fmapdelN _ (fmapdelN _ M a) a0).
intros; intro; intros.
unfold fmapdelN, fmapdel in |- *.
elim (beq_nat_dec l a); intros ee; rewrite ee; auto.
elim (beq_nat_dec l a0); intros ee'; rewrite ee'; auto.
Qed.



Lemma fmapdel_eqf :
 forall B M M' a, eqf M M' -> eqf (fmapdelN B M a) (fmapdelN B M' a).
unfold eqf in |- *; intros.
unfold fmapdelN, fmapdel in |- *.
elim (beq_nat_dec l a); intros ee; rewrite ee; auto.
Qed.



Lemma fmap_listN_fmapdelN_commut :
 forall B Ls M M', eqf M M' -> fmap_listN B Ls M -> fmap_listN B Ls M'.
induction Ls.
compute in |- *; fold beq_nat in |- *; intros.
replace (M' a) with (M a); auto.

induction  a as (a, b); unfold eqf, fmap_listN, fmap_list, fmaplook in |- *;
 fold fmap_list in |- *; IntrosSplitH.
split; auto.
replace (M' a) with (M a); auto.
apply IHLs with (fmapdelN _ M a); auto.

change (eqf (fmapdelN B M a) (fmapdelN B M' a)) in |- *.
apply fmapdel_eqf; auto.
Qed.
Hint Immediate fmapdelN_commut_eq.

Lemma fmap_listN_fmapdel :
 forall B Ls M a,
 fmap_listN B Ls (fmapdelN B M a) -> ~ indomf M a -> fmap_listN B Ls M.
induction Ls; intros.
unfold fmap_listN, fmap_list, nulldomf, notindomf, fmapdelN, fmapdel in H.
unfold fmap_listN, fmap_list, nulldomf, notindomf, fmapdelN, fmapdel in |- *;
 intros.
generalize (H a0); intros.
compare a0 a; intros.
  MCRewrite a0; generalize (not_indom_notindom H0); intros; auto.
  generalize (beq_neq_false _ _ n); intros ee; rewrite <- ee in H1; auto.

induction  a as (a, b).
split.
unfold fmap_listN, fmap_list, nulldomf, notindomf, fmapdelN, fmapdel in H.
inversion_clear H.
UnfoldFmaps.
assert (a <> a0); intros.
   intro.
   clear H2.
   rewrite H in H1; MCClear a.
   generalize (beq_nat_refl a0); intros ee; rewrite <- ee in H1; AbsurdEq.
generalize (beq_neq_false _ _ H); intros.
rewrite <- H3 in H1; auto.

unfold fmap_listN at 1, fmap_list, nulldomf, notindomf in H.
inversion_clear H.
cut (fmap_listN _ Ls (fmapdelN B (fmapdelN B M a) a0)); intros.

assert (~ indomf (fmapdelN _ M a) a0).
   unfold indomf, fmaplook, fmapdelN, fmapdel in |- *; intro.
   elim (beq_nat_dec a0 a); intros ee; rewrite ee in H3.
   MCInv ex; AbsurdEq.
   generalize (H0 H3); auto.

generalize (IHLs _ _ H H3); intros; auto.
apply fmap_listN_fmapdelN_commut with (fmapdelN B (fmapdelN B M a0) a); auto.
Qed.


Lemma limit_ex_fmaplist :
 forall (Lim : nat) B (M : fmap nat B),
 (forall l, Lim <= l -> notindomf M l) -> exists Ls : _, fmap_listN _ Ls M.
induction Lim; unfold notindomf in |- *; intros.
(* Case: Lim = 0 *)
exists (nil (A:=nat * B)).
compute in |- *; intros; auto.
apply H; try apply le_O_n; auto.

(* Case: Lim = (S Lim') *)
assert (forall l, Lim <= l -> notindomf (fmapdelN _ M Lim) l); intros.
  elim (le_lt_or_eq _ _ H0); intros.
  generalize (lt_neq _ _ H1); intros; assert (l <> Lim); auto.
  unfold notindomf, fmapdelN, fmapdel in |- *;
   replace (beq_nat l Lim) with false; try apply beq_neq_false; 
   auto.
  unfold notindomf, fmapdelN, fmapdel in |- *;
   replace (beq_nat l Lim) with true; try MCRewrite Lim;
   try apply beq_nat_refl; auto.
elim (IHLim _ _ H0); intros Ls; intros.
elim (indomf_dec M Lim); intros.
ElimEx H2 b; UnfoldFmaps.
exists ((Lim, b) :: Ls).
split; auto.

exists Ls.
apply fmap_listN_fmapdel with Lim; auto.
Qed.



Lemma fmap_listN_del :
 forall (B : Set) (a : nat) (b : B) M Ls,
 fmap_listN _ ((a, b) :: Ls) M -> fmap_listN _ Ls (fmapdelN _ M a).
unfold fmap_listN at 1, fmap_list in |- *.
IntrosSplitH.
Qed.




Lemma fmap_listN_cons_fmaplook :
 forall B Ls a b M, fmap_listN B ((a, b) :: Ls) M -> fmaplook M a b.
unfold fmap_listN, fmaplook, fmap_list in |- *; IntrosSplitH.
Qed.

Lemma fmaplook_neq_fmapdelN_fmaplook :
 forall B a a' b M,
 a <> a' -> fmaplook (fmapdelN B M a') a b -> fmaplook M a b.
unfold fmaplook, fmapdelN, fmapdel in |- *; intros.
generalize H0; replace (beq_nat a a') with false; try apply beq_neq_false;
 auto; intros.
Qed.


Lemma eqdomf_fmapdelN :
 forall B C M M' a,
 eqdomf nat B C M M' -> eqdomf _ _ _ (fmapdelN _ M a) (fmapdelN _ M' a).
unfold eqdomf, fmapdelN, fmapdel, iff, indomf, fmaplook in |- *; split;
 IntrosSplitH; elim (H a0); intros; elim (beq_nat_dec a0 a); 
 intros ee; rewrite ee in H0; ElimEx H0 bb; try AbsurdEq; 
 MCRewriteRev false; IAssumption.
Qed.
