(* 
  Coq code for the paper:
  Interfacing Hoare Logic and Type Systems for Foundational Proof-Carrying Code
  by Nadeem Abdul Hamid and Zhong Shao

  Coq version 8.0

  This code is a prototype version only -- proofs are mostly complete.
*)



(***************************************************************************)
(* Utilities for manipulating natural numbers *)
(***************************************************************************)

Require Import Arith.
Require Import Div2.
Require Import EqNat.
Require Import Lt.
Require Import Le.



(***************************************************************************)
Fixpoint blt_nat (n m : nat) {struct n} : bool :=
  match n, m with
  | O, O => false
  | O, S _ => true
  | S _, O => false
  | S n1, S m1 => blt_nat n1 m1
  end.


Lemma blt_S_false : forall a : nat, blt_nat (S a) a = false.
Proof.
simple induction a.
unfold blt_nat in |- *.
auto.
intros.
unfold blt_nat in |- *.
auto.
Qed.

Lemma blt_S_eq :
 forall (a b : nat) (t : bool), blt_nat a b = t -> blt_nat (S a) (S b) = t.
Proof.
intros.
unfold blt_nat in |- *.
fold blt_nat in |- *.
auto.
Qed.

Lemma blt_lt_true : forall a b : nat, a < b -> blt_nat a b = true.
Proof.
simple induction a.
intros.
inversion H.
unfold blt_nat in |- *.
auto.
unfold blt_nat in |- *.
auto.
intros.
inversion H0.
apply blt_S_eq.
apply H.
apply lt_n_Sn.
apply blt_S_eq.
apply H.
rewrite <- H2 in H0.
apply lt_S_n.
auto.
Qed.


Lemma blt_Sm_false :
 forall a b : nat, blt_nat a b = false -> blt_nat (S a) b = false.
Proof.
induction a; induction b.
(* Double Induction 1 2. *)
intros.
unfold blt_nat in |- *.
auto.
intros.
unfold blt_nat in H.
absurd (true = false).
discriminate.
auto.
intros.
unfold blt_nat in |- *.
auto.
intros.
unfold blt_nat in H.
fold blt_nat in H.
cut (blt_nat (S a) b = false).
intros.
apply blt_S_eq.
auto.
apply IHa; auto.
Qed.

Lemma blt_false : forall a b : nat, a < b -> blt_nat b a = false.
Proof.
intros.
induction  H as [| m H HrecH].
apply blt_S_false.
apply blt_Sm_false.
auto.
Qed.


Lemma blt_eq_false : forall a : nat, blt_nat a a = false.
Proof.
simple induction a.
unfold blt_nat in |- *.
auto.
intros.
unfold blt_nat in |- *.
auto.
Qed.


Lemma le_blt_false : forall a b : nat, a <= b -> blt_nat b a = false.
Proof.
intros.
induction  H as [| m H HrecH].
apply blt_eq_false.
apply blt_false.
unfold lt in |- *.
apply le_n_S.
auto.
Qed.




(***************************************************************************)


Lemma beq_neq_false : forall a b : nat, a <> b -> false = beq_nat a b.
Proof.
induction a; induction b.
intros.
absurd (0 <> 0); auto.
intros.
unfold beq_nat in |- *.
auto.
intros.
unfold beq_nat in |- *.
auto.
intros.
unfold beq_nat in |- *.
fold beq_nat in |- *.
apply IHa.
auto.
Qed.


Lemma beq_nat_dec :
 forall a b : nat, beq_nat a b = true \/ beq_nat a b = false.
intros; compare a b; intros.
rewrite e; left; auto.
symmetry  in |- *; apply beq_nat_refl.
right; symmetry  in |- *; apply beq_neq_false; auto.
Qed.


Lemma eq_neq_Sl : forall a b : nat, a = b -> S a <> b.
Proof.
induction a; induction b.
intros.
auto.
intros.
absurd (0 = S b).
discriminate.
auto.
intros.
absurd (0 = S a).
discriminate.
auto.
intros.
apply not_eq_S.
auto.
Qed.


(***************************************************************************)
(* utilities for decoding *)

(* to get  n % d  call this as (modaux n d d O), the second argument is
   decremented in recursive calls, the fourth is incremented *)
Fixpoint modaux (n c d r : nat) {struct n} : nat :=
  match n with
  | O => r
  | S p =>
      match c with
      | O => 0
      | S O => modaux p d d 0
      | S c' => modaux p c' d (S r)
      end
  end.

Definition mod_ (n d : nat) : nat := modaux n d d 0.

Definition div4 (n : nat) := div2 (div2 n).
Definition div8 (n : nat) := div4 (div2 n).
Definition div16 (n : nat) := div2 (div8 n).
Definition div32 (n : nat) := div2 (div16 n).


Definition mult4 (n : nat) := 4 * n.
Definition mult8 (n : nat) := 8 * n.
Definition mult16 (n : nat) := 16 * n.
Definition mult32 (n : nat) := 32 * n.


(***************************************************************************)
Lemma eqorneq_nat : forall a a' : nat, a = a' \/ a <> a'.
  intros; compare a a'; auto. Qed.

Lemma le_or_gt : forall m n : nat, n <= m \/ n > m.
induction m; induction n; compute in |- *; auto; intros.
right; apply le_n_S; apply le_O_n.
left; apply le_O_n.
cut (n <= m \/ n > m);
 [ intro HH; elim HH; [ left | right ]; apply le_n_S; auto
 | unfold gt, lt in |- *; auto ].
elim (IHm n); auto.
Qed.

Lemma lt_neq : forall m n : nat, n < m -> n <> m.
induction m;
 [ intros; try match goal with
               | H:(_ < 0) |- _ => inversion H
               end
 | induction n; intros; auto ].
apply not_eq_S; apply IHm; apply lt_S_n; auto.
Qed.

Lemma lt_S_neq : forall n x : nat, n < S x -> x <> n -> n < x.
Proof.
induction n; induction x.
intros.
absurd (0 <> 0); auto.
intros.
apply lt_O_Sn.
intros; unfold lt in H0.
cut (S n <= 0).
intros.
absurd (S n <= 0).
apply le_Sn_O.
auto.
apply le_S_n.
auto.
intros.
apply lt_n_S.
cut (n = x \/ n <> x).
intros.
inversion_clear H1.
absurd (S x <> S x).
auto.
rewrite <- H2 in H0.
auto.
cut (n < S x).
intros.
apply (IHn x H1).
auto.
apply lt_S_n.
auto.
apply eqorneq_nat.
Qed.


Lemma lt_S_eq_or_lt : forall n x : nat, n < S x -> x = n \/ n < x.
Proof.
intros.
cut (x = n \/ x <> n).
intros.
inversion_clear H0.
left; auto.
right; apply lt_S_neq.
auto.
auto.
apply eqorneq_nat.
Qed.


Lemma plus_n_O_n : forall n : nat, n + 0 = n.
Proof.
simple induction n.
unfold plus in |- *.
auto.
intros.
unfold plus in |- *.
fold plus in |- *.
rewrite H.
auto.
Qed.


Lemma plus_S_neq : forall n m : nat, n <> n + S m.
Proof.
simple induction n.
intros.
unfold plus in |- *.
auto.
intros.
unfold plus in |- *.
fold plus in |- *.
auto.
Qed.


(***************************************************************************)