(* 
 *
 * Coq code for the paper
 *      Interfacing Hoare Logic and Type Systems for FPCC
 *         Nadeem Abdul Hamid, 
 *
 * (For Coq version 8.0)
 *
 *)

Require Import EqNat.
Require Import natutil.

(***************************************************************************)
(* Syntax *)

Load "tisreg".  (* Inductive Regt : Set := rA | rB ... | rP *)

Definition Word := nat.
Definition Mem := Word -> Word.
Definition RFile := Regt -> Word.
Definition State := (Mem * (RFile * Word))%type.

Inductive Cmd : Set :=
  | _add : Regt -> Regt -> Regt -> Cmd
  | _addi : Regt -> Regt -> Word -> Cmd
  | _sub : Regt -> Regt -> Regt -> Cmd
  | _subi : Regt -> Regt -> Word -> Cmd
  | _mov : Regt -> Regt -> Cmd
  | _movi : Regt -> Word -> Cmd
  | _bgt : Regt -> Regt -> Word -> Cmd
  | _bgti : Regt -> Word -> Word -> Cmd
  | _jd : Word -> Cmd
  | _jmp : Regt -> Cmd
  | _ld : Regt -> Regt -> Word -> Cmd
  | _st : Regt -> Word -> Regt -> Cmd
  | _ill : Cmd.

(***************************************************************************)
(* decode utilities *)
Definition icoded (w : Word) : nat := mod_ w 8.
Definition r1argd (w : Word) : nat := mod_ (div8 w) 32.
Definition r2argd (w : Word) : nat := mod_ (div32 (div8 w)) 32.
Definition r3argd (w : Word) : nat := mod_ (div32 (div32 (div8 w))) 32.
Definition w1argd (w : Word) : nat := div8 w.
Definition w2argd (w : Word) : nat := div32 (div8 w).
Definition w3argd (w : Word) : nat := div32 (div32 (div8 w)).


(* encode utilities *)
Definition icodee (n : nat) : Word := n.
Definition r1arge (n : nat) : Word := mult8 n.
Definition r2arge (n : nat) : Word := mult32 (mult8 n).
Definition r3arge (n : nat) : Word := mult32 (mult32 (mult8 n)).
Definition w1arge (n : nat) : Word := mult8 n.
Definition w2arge (n : nat) : Word := mult32 (mult8 n).
Definition w3arge (n : nat) : Word := mult32 (mult32 (mult8 n)).


(* Decode function *)
Definition Dc (w : Word) : Cmd :=
  match icoded w with
  | S O =>
      _add (nat2regt (r1argd w)) (nat2regt (r2argd w)) (nat2regt (r3argd w))
  | S (S O) => _addi (nat2regt (r1argd w)) (nat2regt (r2argd w)) (w3argd w)
  | S (S (S O)) =>
      _sub (nat2regt (r1argd w)) (nat2regt (r2argd w)) (nat2regt (r3argd w))
  | S (S (S (S O))) =>
      _subi (nat2regt (r1argd w)) (nat2regt (r2argd w)) (w3argd w)
  | S (S (S (S (S O)))) => _mov (nat2regt (r1argd w)) (nat2regt (r2argd w))
  | S (S (S (S (S (S O))))) => _movi (nat2regt (r1argd w)) (w2argd w)
  | S (S (S (S (S (S (S O)))))) =>
      _bgt (nat2regt (r1argd w)) (nat2regt (r2argd w)) (w3argd w)
  | S (S (S (S (S (S (S (S O))))))) =>
      _bgti (nat2regt (r1argd w)) (w2argd w) (w3argd w)
  | S (S (S (S (S (S (S (S (S O)))))))) => _jd (w1argd w)
  | S (S (S (S (S (S (S (S (S (S O))))))))) => _jmp (nat2regt (r1argd w))
  | S (S (S (S (S (S (S (S (S (S (S O)))))))))) =>
      _ld (nat2regt (r1argd w)) (nat2regt (r2argd w)) (w3argd w)
  | S (S (S (S (S (S (S (S (S (S (S (S O))))))))))) =>
      _st (nat2regt (r1argd w)) (w3argd w) (nat2regt (r2argd w))
  | _ => _ill
  end.


(* Encode function *)
Definition Ec (i : Cmd) : Word :=
  match i with
  | _add rd rs rs' =>
      icodee 1 +
      (r1arge (regt2nat rd) + (r2arge (regt2nat rs) + r3arge (regt2nat rs')))
  | _addi rd rs w =>
      icodee 2 + (r1arge (regt2nat rd) + (r2arge (regt2nat rs) + w3arge w))
  | _sub rd rs rs' =>
      icodee 3 +
      (r1arge (regt2nat rd) + (r2arge (regt2nat rs) + r3arge (regt2nat rs')))
  | _subi rd rs w =>
      icodee 4 + (r1arge (regt2nat rd) + (r2arge (regt2nat rs) + w3arge w))
  | _mov rd rs => icodee 5 + (r1arge (regt2nat rd) + r2arge (regt2nat rs))
  | _movi rd w => icodee 6 + (r1arge (regt2nat rd) + w2arge w)
  | _bgt rd rs w =>
      icodee 7 + (r1arge (regt2nat rd) + (r2arge (regt2nat rs) + w3arge w))
  | _bgti rd i w => icodee 8 + (r1arge (regt2nat rd) + (w2arge i + w3arge w))
  | _jd w => icodee 9 + w1arge w
  | _jmp r => icodee 10 + r1arge (regt2nat r)
  | _ld rd rs w =>
      icodee 11 + (r1arge (regt2nat rd) + (r2arge (regt2nat rs) + w3arge w))
  | _st rd w rs =>
      icodee 12 + (r1arge (regt2nat rd) + (r2arge (regt2nat rs) + w3arge w))
  | _ill => 0
  end.


(***************************************************************************)
(* Machine semantics *)
Definition updatereg (R : RFile) (rd : Regt) (v : Word) : RFile :=
  fun r : Regt => if beq_regt r rd then v else R r.

Definition updatemem (M : Mem) (a v : Word) : Mem :=
  fun w : Word => if beq_nat w a then v else M w.

Definition Step (St : State) : State :=
  match St with
  | (M, (R, pc)) =>
      match Dc (M pc) with
      | _add rd rs rs' => (M, (updatereg R rd (R rs + R rs'), S pc))
      | _addi rd rs w => (M, (updatereg R rd (R rs + w), S pc))
      | _sub rd rs rs' => (M, (updatereg R rd (R rs - R rs'), S pc))
      | _subi rd rs w => (M, (updatereg R rd (R rs - w), S pc))
      | _mov rd rs => (M, (updatereg R rd (R rs), S pc))
      | _movi rd w => (M, (updatereg R rd w, S pc))
      | _bgt rs rt w =>
          if blt_nat (R rt) (R rs) then (M, (R, w)) else (M, (R, S pc))
      | _bgti rs i w =>
          if blt_nat i (R rs) then (M, (R, w)) else (M, (R, S pc))
      | _jd w => (M, (R, w))
      | _jmp r => (M, (R, R r))
      | _ld rd rs w => (M, (updatereg R rd (M (R rs + w)), S pc))
      | _st rd w rs => (updatemem M (R rd + w) (R rs), (R, S pc))
      | _ill => St
      end
  end.

Fixpoint MultiStep (n : nat) : State -> State :=
  match n with
  | O => fun S : State => S(*Step*) 
  | S m => fun S : State => Step (MultiStep m S)
  end.

(***************************************************************************)





